
__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os
import stat
import time

import threading          as mt

import radical.utils      as ru

from ... import states    as rps
from ... import agent     as rpa
from ... import constants as rpc
from ... import utils     as rpu


# ------------------------------------------------------------------------------
# 'enum' for RP's spawner types
EXECUTING_NAME_POPEN   = 'POPEN'
EXECUTING_NAME_FLUX    = 'FLUX'
EXECUTING_NAME_SLEEP   = 'SLEEP'
EXECUTING_NAME_DRAGON  = 'DRAGON'


# ------------------------------------------------------------------------------
#
class AgentExecutingComponent(rpu.AgentComponent):
    '''
    Manage the creation of Task processes, and watch them until they are
    completed (one way or the other).  The spawner thus moves the task from
    PendingExecution to Executing, and then to a final state (or PendingStageOut
    of course).
    '''
    _shell     = ru.which('bash') or '/bin/sh'
    _header    = '#!%s\n' % _shell
    _separator = '\n# ' + '-' * 78 + '\n'


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Spawner
    #
    @classmethod
    def create(cls, cfg, session):

        # Make sure that we are the base-class!
        if cls != AgentExecutingComponent:
            raise TypeError('Factory only available to base class!')

        name = session.rcfg.agent_spawner

        from .popen    import Popen
        from .flux     import Flux
        from .sleep    import Sleep
        from .dragon   import Dragon

        impl = {
            EXECUTING_NAME_POPEN : Popen,
            EXECUTING_NAME_FLUX  : Flux,
            EXECUTING_NAME_SLEEP : Sleep,
            EXECUTING_NAME_DRAGON: Dragon,
        }

        if name not in impl:
            raise ValueError('AgentExecutingComponent %s unknown' % name)

        return impl[name](cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):


        rm_name  = self.session.rcfg.resource_manager
        self._rm = rpa.ResourceManager.create(rm_name,
                                              self.session.cfg,
                                              self.session.rcfg,
                                              self._log, self._prof)

        self._pwd      = os.path.realpath(os.getcwd())
        self.sid       = self.session.uid
        self.pid       = self.session.cfg.pid
        self.resource  = self.session.cfg.resource
        self.rsbox     = self.session.cfg.resource_sandbox
        self.ssbox     = self.session.cfg.session_sandbox
        self.psbox     = self.session.cfg.pilot_sandbox
        self.gtod      = '$RP_PILOT_SANDBOX/gtod'
        self.prof      = '$RP_PILOT_SANDBOX/prof'

        # if so configured, let the tasks know what to use as tmp dir
        self._task_tmp = self.session.rcfg.get('task_tmp',
                                               os.environ.get('TMP', '/tmp'))

        if self.psbox.startswith(self.ssbox):
            self.psbox = '$RP_SESSION_SANDBOX%s'  % self.psbox[len(self.ssbox):]
        if self.ssbox.startswith(self.rsbox):
            self.ssbox = '$RP_RESOURCE_SANDBOX%s' % self.ssbox[len(self.rsbox):]
        if self.ssbox.endswith(self.sid):
            self.ssbox = '%s$RP_SESSION_ID/'      % self.ssbox[:-len(self.sid)]

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)

        self._to_tasks  = list()
        self._to_lock   = mt.Lock()
        self._to_thread = mt.Thread(target=self._to_watcher)
        self._to_thread.daemon = True
        self._to_thread.start()


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        raise NotImplementedError('work is not implemented')


    # --------------------------------------------------------------------------
    #
    def control_cb(self, topic, msg):

        self._log.info('command_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']
        arg = msg['arg']

        # FIXME RPC: already handled in the component base class
        if cmd == 'cancel_tasks':

            self._log.info('cancel_tasks command (%s)', arg)
            for tid in arg['uids']:
                self.cancel_task(tid)


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, uid):

        raise NotImplementedError('cancel_task is not implemented')


    # --------------------------------------------------------------------------
    #
    def _to_watcher(self):
        '''
        watch the set of tasks for which timeouts are defined.  If the timeout
        passes and the tasks are still active, kill the task via
        `self._cancel_task(task)`.  That has to be implemented by al executors.
        '''

        while True:

            # check once per second at most
            time.sleep(1)

            now = time.time()
            with self._to_lock:

                # running tasks for next check
                to_list = list()
                for to, start, task in self._to_tasks:
                    if now - start > to:
                        self._prof.prof('task_timeout', uid=task['uid'])
                        self.cancel_task(uid=task['uid'])
                    else:
                        to_list.append([to, start, task])

                self._to_tasks = to_list


    # --------------------------------------------------------------------------
    #
    def handle_timeout(self, task):

        to = task['description'].get('timeout', 0.0)

        if to > 0.0:
            with self._to_lock:
                self._to_tasks.append([to, time.time(), task])


    # --------------------------------------------------------------------------
    #
    def advance_tasks(self, tasks, state, publish, push, ts=None):
        '''
        sort tasks into different buckets, depending on their origin.
        That origin will determine where tasks which completed execution
        and end up here will be routed to:

          - client: state update to update worker
          - raptor: state update to `STATE_PUBSUB`
          - agent : state update to `STATE_PUBSUB`

        a fallback is not in place to enforce the specification of the
        `origin` attributes for tasks.
        '''


        buckets = {'client': list(),
                   'raptor': list(),
                   'agent' : list()}

        for task in ru.as_list(tasks):
            buckets[task['origin']].append(task)

        # we want any task which has a `raptor_id` set to show up in raptor's
        # result callbacks
        if state != rps.AGENT_EXECUTING:
            for task in ru.as_list(tasks):
                if task['description'].get('raptor_id'):
                    if task not in buckets['raptor']:
                        buckets['raptor'].append(task)

        if buckets['client']:
            self.advance(buckets['client'], state=state,
                                            publish=publish, push=push, ts=ts)

        if buckets['raptor']:
            self.advance(buckets['raptor'], state=state,
                                            publish=publish, push=False, ts=ts)
            self.publish(rpc.STATE_PUBSUB, {'cmd': 'raptor_state_update',
                                            'arg': buckets['raptor']})

        if buckets['agent']:
            self.advance(buckets['agent'], state=state,
                                            publish=publish, push=False, ts=ts)


    # --------------------------------------------------------------------------
    #
    # methods to prepare task exec scripts
    #
    def _create_exec_script(self, launcher, task):

        tid  = task['uid']
        td   = task['description']
        sbox = task['task_sandbox_path']

        exec_script   = '%s.exec.sh' % tid
        exec_path     = '$RP_TASK_SANDBOX/%s' % exec_script
        exec_fullpath = '%s/%s' % (sbox, exec_script)

        # make sure the sandbox exists
        self._prof.prof('task_mkdir', uid=tid)
        ru.rec_makedir(sbox)
        self._prof.prof('task_mkdir_done', uid=tid)


        # the exec shell script runs the same set of commands for all ranks.
        # However, if the ranks need different GPU's assigned, or if either pre-
        # or post-exec directives contain per-rank dictionaries, then we switch
        # per-rank in the script for all sections between pre- and post-exec.

        n_ranks = td['ranks']
        slots   = task.setdefault('slots', {})

        self._extend_pre_exec(td, slots.get('ranks'))

        with ru.ru_open(exec_fullpath, 'w') as fout:

            tmp  = ''
            tmp += self._header
            tmp += self._separator
            tmp += self._get_rp_env(task)
            tmp += self._get_rp_funcs()
            tmp += self._separator
            tmp += '# rank ID\n'
            tmp += self._get_rank_ids(n_ranks, launcher)
            tmp += self._separator
            tmp += self._get_prof('exec_start')

            tmp += self._get_task_env(task, launcher)

            tmp += self._separator
            tmp += '# pre-exec commands\n'
            tmp += self._get_prof('exec_pre')
            tmp += self._get_prep_exec(task, n_ranks, sig='pre_exec')

            tmp += self._separator
            tmp += '# execute rank\n'
            tmp += self._get_prof('rank_start')
            tmp += self._get_exec(task, launcher)
            tmp += self._get_prof('rank_stop',
                                  msg='RP_EXEC_PID=$RP_EXEC_PID:'
                                      'RP_RANK_PID=$RP_RANK_PID')

            tmp += self._separator
            tmp += '# post-exec commands\n'
            tmp += self._get_prof('exec_post')
            tmp += self._get_prep_exec(task, n_ranks, sig='post_exec')

            tmp += self._separator
            tmp += self._get_prof('exec_stop')
            tmp += 'exit $RP_RET\n'

            tmp += self._separator
            tmp += '\n'

            fout.write(tmp)

        # make sure scripts are executable
        st_e = os.stat(exec_fullpath)
        os.chmod(exec_fullpath, st_e.st_mode | stat.S_IEXEC)

        # need to set `DEBUG_5` or higher to get slot debug logs
        if self._log._debug_level >= 5:
            ru.write_json('%s/%s.sl' % (sbox, tid), slots)

        return exec_path, exec_fullpath


    # --------------------------------------------------------------------------
    #
    def _get_task_env(self, task, launcher):

        ret = ''
        td  = task['description']

        # named_env's are prepared by the launcher
        if td['named_env']:
            ret += '\n# named environment\n'
            ret += '. %s\n' % launcher.get_task_named_env(td['named_env'])

        # also add any env vars requested in the task description
        if td['environment']:
            ret += '\n# task env settings\n'
            for key, val in td['environment'].items():
                ret += 'export %s="%s"\n' % (key, val)

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rank_ids(self, n_ranks, launcher):

        ret  = ''
        ret += 'export RP_RANKS=%s\n' % n_ranks
        ret += launcher.get_rank_cmd()

        if n_ranks > 1:

            # make sure that RP_RANK is known (otherwise task fails silently)
            if 'export RP_RANK=' not in ret:
                raise RuntimeError('launch method does not export RP_RANK')

        # also define a method to sync all ranks on certain events
        ret += '\nrp_sync_ranks() {\n'
        ret += '    sig=$1\n'
        ret += '    echo $RP_RANK >> $sig.sig\n'
        ret += '    while test $(cat $sig.sig | wc -l) -lt $RP_RANKS; do\n'
        ret += '        sleep 1\n'
        ret += '    done\n'
        ret += '}\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _extend_pre_exec(self, td, ranks=None):

        # FIXME: this assumes that the rank has a `gpu_maps` and `core_maps`
        #        with exactly one entry, corresponding to the rank process to be
        #        started.

        # FIXME: need to distinguish between logical and physical IDs

        if td['threading_type'] == rpc.OpenMP:
            # for future updates: if task ranks are heterogeneous in terms of
            #                     number of threads, then the following string
            #                     should be converted into dictionary (per rank)
            num_threads = td.get('cores_per_rank', 1)
            td['pre_exec'].append('export OMP_NUM_THREADS=%d' % num_threads)

        if td['gpus_per_rank'] and td['gpu_type'] == rpc.CUDA and ranks:
            # equivalent to the 'physical' value for original `cvd_id_mode`
            rank_id  = 0
            rank_env = {}
            for slot_ranks in ranks:
                for gpu_map in slot_ranks['gpu_map']:
                    rank_env[str(rank_id)] = \
                        'export CUDA_VISIBLE_DEVICES=%s' % \
                        ','.join([str(g) for g in gpu_map])
                    rank_id += 1
            td['pre_exec'].append(rank_env)

        # pre-defined `pre_exec` per platform configuration
        td['pre_exec'].extend(ru.as_list(self.session.rcfg.get('task_pre_exec')))


    # --------------------------------------------------------------------------
    #
    def _get_prep_exec(self, task, n_ranks, sig):

        ret = ''
        td  = task['description']

        if sig not in td:
            return ret

        entries         = ru.as_list(td[sig])
        switch_per_rank = any([isinstance(x, dict) for x in entries])
        cmd_template    = '%s || rp_error %s\n'

        sync_ranks_cmd = ''
        if sig == 'pre_exec' and td['pre_exec_sync']:
            sync_ranks_cmd = 'rp_sync_ranks %s\n' % sig

        if not switch_per_rank:
            return ''.join([cmd_template % (x, sig) for x in entries]) + \
                   sync_ranks_cmd

        ret += 'case "$RP_RANK" in\n'
        for rank_id in range(n_ranks):

            ret += '    %d)\n' % rank_id

            for entry in entries:

                if isinstance(entry, str):
                    entry = {str(rank_id): entry}

                for cmd in ru.as_list(entry.get(str(rank_id))):
                    ret += '        ' + cmd_template % (cmd, sig)

            ret += '        ;;\n'

        ret += 'esac\n' + sync_ranks_cmd

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_exec(self, task, launcher):

        # FIXME: core pinning goes here

        ret  = '%s &\n' % launcher.get_exec(task)

        # collect PIDs for exec-script and executable
        ret += '\nRP_EXEC_PID=$$\nRP_RANK_PID=$!\n\n'
        ret += 'wait $RP_RANK_PID\n'

        # set output
        ret += 'RP_RET=$?\n'

        return ret



    # --------------------------------------------------------------------------
    #
    # methods to prepare task launch scripts
    #
    def _create_launch_script(self, launcher, task, exec_path):

        tid  = task['uid']
        sbox = task['task_sandbox_path']

        if not launcher:
            raise RuntimeError('no launcher found for task %s' % tid)

        self._log.debug('Launching task with %s', launcher.name)

        launch_script   = '%s.launch.sh' % tid
        launch_path     = '$RP_TASK_SANDBOX/%s' % launch_script
        launch_fullpath = '%s/%s' % (sbox, launch_script)

        ru.rec_makedir(sbox)

        with ru.ru_open(launch_fullpath, 'w') as fout:

            tmp  = ''
            tmp += self._header
            tmp += self._separator
            tmp += self._get_rp_env(task)
            tmp += self._get_rp_funcs()
            tmp += self._separator
            tmp += self._get_prof('launch_start')

            tmp += self._separator
            tmp += '# change to task sandbox\n'
            tmp += 'cd $RP_TASK_SANDBOX\n'

            tmp += self._separator
            tmp += '# prepare launcher env\n'
            tmp += self._get_launch_env(launcher)

            tmp += self._separator
            tmp += '# pre-launch commands\n'
            tmp += self._get_prof('launch_pre')
            tmp += self._get_prep_launch(task, sig='pre_launch')

            tmp += self._separator
            tmp += '# launch commands\n'
            tmp += self._get_prof('launch_submit')
            tmp += self._get_launch(task, launcher, exec_path)
            tmp += self._get_prof('launch_collect',
                                  msg='RP_LAUNCH_PID=$RP_LAUNCH_PID')

            tmp += self._separator
            tmp += '# post-launch commands\n'
            tmp += self._get_prof('launch_post')
            tmp += self._get_prep_launch(task, sig='post_launch')

            tmp += self._separator
            tmp += self._get_prof('launch_stop')
            tmp += 'exit $RP_RET\n'

            tmp += self._separator
            tmp += '\n'

            fout.write(tmp)


        st_l = os.stat(launch_fullpath)
        os.chmod(launch_fullpath, st_l.st_mode | stat.S_IEXEC)

        return launch_path, launch_fullpath


    # --------------------------------------------------------------------------
    #
    def _get_launch(self, task, launcher, exec_path):

        ret  = '( \\\n'

        for cmd in ru.as_list(launcher.get_launch_cmds(task, exec_path)):
            ret += '  %s \\\n' % cmd

        ret += ') 1> %s \\\n  2> %s\n' % (task['stdout_file_short'],
                                          task['stderr_file_short'])
        # collect PID for launch-script
        ret += 'RP_RET=$?\n'
        ret += 'RP_LAUNCH_PID=$$\n'

        return ret


    # --------------------------------------------------------------------------
    #
    # launcher
    #
    def _get_launch_env(self, launcher):

        ret  = ''

        for cmd in launcher.get_launcher_env():
            ret += '%s || rp_error launcher_env\n' % cmd

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_prep_launch(self, task, sig):

        ret = ''
        td = task['description']

        if sig not in td:
            return ret

        for cmd in ru.as_list(task['description'][sig]):
            ret += '%s || rp_error %s\n' % (cmd, sig)

        return ret


    # --------------------------------------------------------------------------
    #
    # methods to prepare task launch and exec scripts
    #
    def _get_rp_env(self, task):

        tid  = task['uid']
        td   = task['description']
        name = task.get('name') or tid
        sbox = os.path.realpath(task['task_sandbox_path'])

        if sbox.startswith(self._pwd):
            sbox = '$RP_PILOT_SANDBOX%s' % sbox[len(self._pwd):]

        gpr = td['gpus_per_rank']
        if int(gpr) == gpr:
            gpr = '%d' % gpr
        else:
            gpr = '%f' % gpr

        ret  = '\n'
        ret += 'export RP_TASK_ID="%s"\n'          % tid
        ret += 'export RP_TASK_NAME="%s"\n'        % name
        ret += 'export RP_PILOT_ID="%s"\n'         % self.pid
        ret += 'export RP_SESSION_ID="%s"\n'       % self.sid
        ret += 'export RP_RESOURCE="%s"\n'         % self.resource
        ret += 'export RP_RESOURCE_SANDBOX="%s"\n' % self.rsbox
        ret += 'export RP_SESSION_SANDBOX="%s"\n'  % self.ssbox
        ret += 'export RP_PILOT_SANDBOX="%s"\n'    % self.psbox
        ret += 'export RP_TASK_SANDBOX="%s"\n'     % sbox
        ret += 'export RP_REGISTRY_ADDRESS="%s"\n' % self.session.reg_addr
        ret += 'export RP_CORES_PER_RANK=%d\n'     % td['cores_per_rank']
        ret += 'export RP_GPUS_PER_RANK=%s\n'      % gpr

        # FIXME AM
      # ret += 'export RP_LFS="%s"\n'              % self.lfs
        ret += 'export RP_GTOD="%s"\n'             % self.gtod
        ret += 'export RP_PROF="%s"\n'             % self.prof

        if self._prof.enabled:
            ret += 'export RP_PROF_TGT="%s/%s.prof"\n' % (sbox, tid)
        else:
            ret += 'unset  RP_PROF_TGT\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rp_funcs(self):

        # define helper functions
        ret  = '\nrp_error() {\n'
        ret += '    echo "$1 failed" 1>&2\n'
        ret += '    exit 1\n'
        ret += '}\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_prof(self, event, msg=''):

        return '$RP_PROF %s "%s"\n' % (event, msg)



# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class Slurm(ResourceManager):

    # --------------------------------------------------------------------------
    #
    @staticmethod
    def batch_started():

        return bool(os.getenv('SLURM_JOB_ID'))

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        nodelist = os.environ.get('SLURM_NODELIST') or \
                   os.environ.get('SLURM_JOB_NODELIST')
        if nodelist is None:
            raise RuntimeError('$SLURM_*NODELIST not set')

        # Parse SLURM nodefile environment variable
        node_names = ru.get_hostlist(nodelist)
        self._log.info('found nodelist %s. Expanded to: %s',
                       nodelist, node_names)

        if not rm_info.cores_per_node:
            # $SLURM_CPUS_ON_NODE = Number of physical cores per node
            cpn_str = os.environ.get('SLURM_CPUS_ON_NODE')
            if cpn_str is None:
                raise RuntimeError('$SLURM_CPUS_ON_NODE not set')
            rm_info.cores_per_node = int(cpn_str)

        if not rm_info.gpus_per_node:
            if os.environ.get('SLURM_GPUS_ON_NODE'):
                rm_info.gpus_per_node = int(os.environ['SLURM_GPUS_ON_NODE'])
            else:
                # GPU IDs per node
                # - global context: SLURM_JOB_GPUS and SLURM_STEP_GPUS
                # - cgroup context: GPU_DEVICE_ORDINAL
                gpu_ids = os.environ.get('SLURM_JOB_GPUS')  or \
                          os.environ.get('SLURM_STEP_GPUS') or \
                          os.environ.get('GPU_DEVICE_ORDINAL')
                if gpu_ids:
                    rm_info.gpus_per_node = len(gpu_ids.split(','))

        nodes = [(node, rm_info.cores_per_node) for node in node_names]

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        return rm_info


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2024, http://radical.rutgers.edu"
__license__   = "MIT"


# ------------------------------------------------------------------------------
# we *first* import radical.utils, so that the monkeypatching of the logger has
# a chance to kick in before the logging module is pulled by any other 3rd party
# module, and also to monkeypatch `os.fork()` for the `atfork` functionality
#
import os            as _os
import radical.utils as _ru


# ------------------------------------------------------------------------------
# constants and types
from .states     import *
from .constants  import *


# ------------------------------------------------------------------------------
# import API
from .session                   import Session
from .proxy                     import Proxy

from .task_manager              import TaskManager
from .task                      import Task
from .raptor_tasks              import RaptorMaster, RaptorWorker
from .pytask                    import PythonTask
from .task_description          import TaskDescription
from .task_description          import TASK_EXECUTABLE
from .task_description          import TASK_METH, TASK_METHOD
from .task_description          import TASK_FUNC, TASK_FUNCTION
from .task_description          import TASK_EXEC, TASK_EVAL
from .task_description          import TASK_PROC, TASK_SHELL
from .task_description          import RAPTOR_MASTER, RAPTOR_WORKER
from .task_description          import AGENT_SERVICE
from .resource_config           import ResourceConfig

from .pilot_manager             import PilotManager
from .pilot                     import Pilot
from .pilot_description         import PilotDescription

pythontask = PythonTask.pythontask


# ------------------------------------------------------------------------------
# make submodules available -- mostly for internal use
from . import utils
from . import tmgr
from . import pmgr
from . import agent

from .agent  import Agent_0
from .agent  import Agent_n

from .raptor import Master, Worker


# ------------------------------------------------------------------------------
#
# get version info
#
_mod_root = _os.path.dirname (__file__)

version_short, version_base, version_branch, version_tag, version_detail \
             = _ru.get_version(_mod_root)
version      = version_short
__version__  = version_detail



# ------------------------------------------------------------------------------





import sys
import time
import queue

import threading       as mt
import multiprocessing as mp
import radical.utils   as ru


_TIMEOUT         =   300  # time to keep the bridge alive
_LINGER_TIMEOUT  =   250  # ms to linger after close
_HIGH_WATER_MARK =     0  # number of messages to buffer before dropping
                          # 0:  infinite


# ------------------------------------------------------------------------------
# This ZMQ bridge links clients and agents, and bridges network gaps.  As such
# it needs to run on a resource which has a public IP address that can be
# reached from both the client and the server machine.
#
# The bridge listens on a `REP` socket (`bridge_request`) for incoming client or
# agent connections, identified by a common session ID.  A client connection
# will trigger the creation of the following communication channels:
#
#   - proxy_control_pubsub_bridge
#     links client and agent control pubsubs (includes heartbeat)
#   - proxy_state_pubsub_bridge
#     forwards task state updates from agents to client
#   - proxy_task_queue
#     forwards tasks from the client to the agents and vice versa
#
#
# The protocol on the `bridge_request` channel is as follows:
#
# register
# --------
#
#    request:
#       'cmd': 'register'
#       'arg': 'sid': <sid>
#
#    reply:
#       'res': {'proxy_control_pubsub': {'sub': <url>, 'pub': <url>},
#               'proxy_state_pubsub'  : {'sub': <url>, 'pub': <url>},
#               'proxy_task_queue'    : {'put': <url>, 'get': <url>}}
#
#    notes:
#      - the request will fail if the session ID is known from another
#        `register` call
#        'err': 'sid already connected'
#      - this request should otherwise always succeed
#      - the created pubsub channels will be terminated if the control channel
#        has not seen a client heartbeat for <10 * heartbeat_interval> seconds
#        - see semantics of the 'unregister' request for details.
#      - the same termination semantics holds for the 'unregister'
#        request.
#      - any task queues which exist for that session at the time of
#        termination will also be closed, disregarding any data held in those
#        queues.
#
#
# lookup
# ------
#
#    request:
#       'cmd': 'lookup'
#       'arg': 'sid': <sid>
#
#    reply:
#       'res': {'proxy_control_pubsub': {'sub': <url>, 'pub': <url>},
#               'proxy_state_pubsub'  : {'sub': <url>, 'pub': <url>},
#               'proxy_task_queue'    : {'put': <url>, 'get': <url>}}
#
#    notes:
#      - the request will fail if the session ID is not registered (anymore)
#      - this request should otherwise always succeed
#      - the call returns the same information as `register`, but does
#        not alter the state of the client's bridge in any other way.
#      - the request does not count as a heartbeat
#
#
# unregister
# ----------
#
#    request:
#       'cmd': 'unregister'
#       'arg': 'sid': <sid>
#
#    reply:
#       'res': 'ok'
#
#   - this method only fails when the session is not connected, with
#     'err': 'session not connected'
#   - in all other cases, the request will cause the immediate termination of
#     all ZMQ bridges (pubsubs and queues) previously created for that
#     session, disregarding of their state, and disposing all undelivered
#     messages still held in the bridges.
#
#
# heartbeat
# ---------
#
#    request:
#      'cmd': 'heartbeat'
#      'arg': 'sid': <sid>
#
#    reply:
#      'res': {'time': <heartbeat timestamp>}
#
# notes:
#    - this request will fail if the session is either not connected or timed
#      because of an earlier heartbeat failure:
#      'err': 'session not connected'
#    - it will otherwise ensure the server that the client is still alive and
#      requires the bridge to be up.  If the server does not receive a heartbeat
#      for longer than TIMEOUT seconds, the bridge will be terminated.
#
#
# default error mode
# ------------------
#
# To any request other than the above, the ZMQ bridge will respond:
#      'err': 'invalid request'
#
# ------------------------------------------------------------------------------

# ------------------------------------------------------------------------------
#
class Proxy(ru.zmq.Server):

    def __init__(self, path=None):

        self._lock    = mt.Lock()
        self._clients = dict()

        ru.zmq.Server.__init__(self, url='tcp://*:10000+', path=path)

        self._monitor_thread = mt.Thread(target=self._monitor)
        self._monitor_thread.daemon = True
        self._monitor_thread.start()

        self.register_request('register',     self._register)
        self.register_request('lookup',       self._lookup)
        self.register_request('unregister',   self._unregister)
        self.register_request('heartbeat',    self._heartbeat)
        self.register_request('service_stop', self._service_stop)


    # --------------------------------------------------------------------------
    #
    def _monitor(self):

        # this is a daemon thread - it never exits until process termination
        while True:

            time.sleep(10)
            now  = time.time()

            # iterate w/o lock, and thus get a snapshot of the known sids
            sids = list(self._clients.keys())

            to_terminate = list()
            for sid in sids:

                client = self._clients.get(sid)
                if not client:
                    continue

                if now > (client['hb'] + _TIMEOUT):
                    self._log.warn('client %s timed out' % sid)
                    to_terminate.append(sid)

            if not to_terminate:
                continue

            with self._lock:

                for sid in to_terminate:

                    client = self._clients.get(sid)
                    if not client:
                        continue

                    client['term'].set()
                    client['proc'].join()
                    del(self._clients[sid])


    # --------------------------------------------------------------------------
    #
    def stop(self):

        for sid in self._clients:
            self._log.info('stop client %s' % sid)
            self._clients[sid]['term'].set()

        self._log.info('stop proxy service')
        ru.zmq.Server.stop(self)


    # --------------------------------------------------------------------------
    #
    def _register(self, arg):

        sid = arg['sid']

        if sid in self._clients:
            raise RuntimeError('client already registered')

        q = mp.Queue()
        term = mp.Event()
        proc = mp.Process(target=self._worker, args=(sid, q, term, self._path))
        proc.start()

        try:
            cfg = q.get(timeout=10)
        except queue.Empty as e:
            proc.terminate()
            raise RuntimeError('worker startup failed') from e

        self._clients[sid] = {'proc': proc,
                              'term': term,
                              'cfg' : cfg,
                              'hb'  : time.time()}

        return self._clients[sid]['cfg']


    # --------------------------------------------------------------------------
    #
    def _worker(self, sid, q, term, path):

        # FIXME: log level etc
        log = ru.Logger('%s.proxy' % sid, level='debug', path=path)

        proxy_cp = None
        proxy_sp = None
        proxy_tq = None

        try:
            proxy_cp = ru.zmq.PubSub(channel='proxy_control_pubsub',
                                     cfg={'uid'    : 'proxy_control_pubsub',
                                          'type'   : 'pubsub',
                                          'log_lvl': 'debug',
                                          'path'   : path})

            proxy_sp = ru.zmq.PubSub(channel='proxy_state_pubsub',
                                     cfg={'uid'    : 'proxy_state_pubsub',
                                          'type'   : 'pubsub',
                                          'log_lvl': 'debug',
                                          'path'   : path})

            proxy_tq = ru.zmq.Queue (channel='proxy_task_queue',
                                     cfg={'uid'    : 'proxy_task_queue',
                                          'type'   : 'queue',
                                          'log_lvl': 'debug',
                                          'path'   : path})

            proxy_cp.start()
            proxy_sp.start()
            proxy_tq.start()

            cfg = {'proxy_control_pubsub': {'addr_pub': str(proxy_cp.addr_pub),
                                            'addr_sub': str(proxy_cp.addr_sub)},
                    'proxy_state_pubsub' : {'addr_pub': str(proxy_sp.addr_pub),
                                            'addr_sub': str(proxy_sp.addr_sub)},
                    'proxy_task_queue'   : {'addr_put': str(proxy_tq.addr_put),
                                            'addr_get': str(proxy_tq.addr_get)}}

            # inform service about endpoint details
            q.put(cfg)

            # we run forever until we receive a termination command
            log.info('work')
            term.wait()


        except:
            log.exception('worker failed')

        finally:

            if proxy_cp: proxy_cp.stop()
            if proxy_sp: proxy_sp.stop()
            if proxy_tq: proxy_tq.stop()

            log.info('terminated')


    # --------------------------------------------------------------------------
    #
    def _lookup(self, arg):

        sid = arg['sid']

        with self._lock:
            if sid not in self._clients:
                raise RuntimeError('client %s not registered' % sid)

            return self._clients[sid]['cfg']


    # --------------------------------------------------------------------------
    #
    def _unregister(self, arg):

        sid = arg['sid']

        with self._lock:

            if sid not in self._clients:
                raise RuntimeError('client %s not registered' % sid)

            self._clients[sid]['term'].set()
            self._clients[sid]['proc'].join()

            del(self._clients[sid])


    # --------------------------------------------------------------------------
    #
    def _heartbeat(self, arg):

        sid = arg['sid']
        now = time.time()

        with self._lock:

            if sid not in self._clients:
                self._log.warn('client %s not in %s', sid, self._clients)
                return

            self._clients[sid]['hb'] = now


    # --------------------------------------------------------------------------
    #
    def _service_stop(self):

        self.stop()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'

import os
import sys

verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

import radical.pilot as rp
import radical.utils as ru


# ------------------------------------------------------------------------------
#
# READ the RADICAL-Pilot documentation: https://radicalpilot.readthedocs.io/
#
# ------------------------------------------------------------------------------


# -----------------------------------------------------------------------------
#
if __name__ == '__main__':

    # we use a reporter class for nicer output
    report = ru.Reporter(name='radical.pilot')
    report.title('Getting Started (RP version %s)' % rp.version)

    # use the resource specified as argument, fall back to localhost
    if len(sys.argv) >= 2  : resources = sys.argv[1:]
    else                   : resources = ['local.localhost']

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:

        # read the config used for resource details
        report.info('read config')
        config = ru.read_json('%s/config.json' % os.path.dirname(os.path.abspath(__file__)))
        report.ok('>>ok\n')

        report.header('submit pilots')

        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        pmgr = rp.PilotManager(session=session)

        # Define an [n]-core local pilot that runs for [x] minutes
        # Here we use a dict to initialize the description object
        pdescs = list()
        for resource in resources:
            pd_init = {
                       'resource'      : resource,
                       'runtime'       : 15,  # pilot runtime (min)
                       'exit_on_error' : True,
                       'project'       : config[resource].get('project', None),
                       'queue'         : config[resource].get('queue', None),
                       'access_schema' : config[resource].get('schema', None),
                       'cores'         : config[resource].get('cores', 1),
                       'gpus'          : config[resource].get('gpus', 0),
                      }
            pdescs.append(rp.PilotDescription(pd_init))

        # Launch the pilots.
        pilots = pmgr.submit_pilots(pdescs)

        # Register the Pilots in a TaskManager object.
        tmgr = rp.TaskManager(session=session)
        tmgr.add_pilots(pilots)

        for gen in range(2):

            report.header('submit tasks [%d]' % gen)

            # Create a workload of Tasks.
            # Each task reports the id of the pilot it runs on.

            n = 128   # number of tasks to run
            report.info('create %d task description(s)\n\t' % n)

            tds = list()
            for i in range(0, n):

                # create a new Task description, and fill it.
                # Here we don't use dict initialization.
                td = rp.TaskDescription()
                td.executable = '/bin/echo'
                td.arguments  = ['$RP_PILOT_ID']

                tds.append(td)
                report.progress()
            report.ok('>>ok\n')
            tasks = tmgr.submit_tasks(tds)
            report.header('gather results')
            tmgr.wait_tasks()

            report.info('\n')
            counts = dict()
            for task in tasks:
                out_str = task.stdout.strip()[:35]
                report.plain('  * %s: %s, exit: %3s, out: %s\n'
                        % (task.uid, task.state[:4],
                            task.exit_code, out_str))
                if out_str not in counts:
                    counts[out_str] = 0
                counts[out_str] += 1

            report.info("\n")
            for out_str in counts:
                report.info("  * %-20s: %3d\n" % (out_str, counts[out_str]))
            report.info("  * %-20s: %3d\n" % ('total', sum(counts.values())))


    except Exception as e:
        # Something unexpected happened in the pilot code above
        report.error('caught Exception: %s\n' % e)
        raise

    except (KeyboardInterrupt, SystemExit):
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        report.warn('exit requested\n')

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.  This will kill all remaining pilots.
        report.header('finalize')
        session.close()

    report.header()


# ------------------------------------------------------------------------------





import io
import os
import sys
import time
import shlex

import threading         as mt

import radical.utils     as ru

from .. import states    as rps
from .. import constants as rpc

from ..pytask           import PythonTask
from ..task_description import TASK_FUNC, TASK_METH, TASK_EXEC
from ..task_description import TASK_PROC, TASK_SHELL, TASK_EVAL


# ------------------------------------------------------------------------------
#
class Worker(object):
    '''
    Implement the Raptor protocol for dispatching multiple Tasks on persistent
    resources.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, manager, rank, raptor_id):

        self._manager   = manager
        self._rank      = rank
        self._raptor_id = raptor_id
        self._reg_event = mt.Event()
        self._reg_addr  = os.environ['RP_REGISTRY_ADDRESS']
        self._sbox      = os.environ['RP_TASK_SANDBOX']
        self._uid       = os.environ['RP_TASK_ID']
        self._sid       = os.environ['RP_SESSION_ID']
        self._ranks     = int(os.environ['RP_RANKS'])

        self._reg       = ru.zmq.RegistryClient(url=self._reg_addr)
        self._cfg       = ru.Config(cfg=self._reg['cfg'])

        self._hb_delay  = self._reg['rcfg.raptor.hb_delay']

        self._log  = ru.Logger(name=self._uid,
                               ns='radical.pilot.worker',
                               level=self._cfg.log_lvl,
                               debug=self._cfg.debug_lvl,
                               targets=self._cfg.log_tgt,
                               path=self._cfg.path)
        self._prof = ru.Profiler(name='%s.%04d' % (self._uid, self._rank),
                                 ns='radical.pilot.worker',
                                 path=self._sbox)

        # register for lifetime management messages on the control pubsub
        psbox     = os.environ['RP_PILOT_SANDBOX']
        state_cfg = self._reg['bridges.%s' % rpc.STATE_PUBSUB]
        ctrl_cfg  = self._reg['bridges.%s' % rpc.CONTROL_PUBSUB]

        ru.zmq.Subscriber(rpc.STATE_PUBSUB, url=state_cfg['addr_sub'],
                          log=self._log, prof=self._prof, cb=self._state_cb,
                          topic=rpc.STATE_PUBSUB)
        ru.zmq.Subscriber(rpc.CONTROL_PUBSUB, url=ctrl_cfg['addr_sub'],
                          log=self._log, prof=self._prof, cb=self._control_cb,
                          topic=rpc.CONTROL_PUBSUB)

        # we push hertbeat and registration messages on that pubsub also
        self._ctrl_pub = ru.zmq.Publisher(rpc.CONTROL_PUBSUB,
                                          url=ctrl_cfg['addr_pub'],
                                          log=self._log,
                                          prof=self._prof)
        # let ZMQ settle
        time.sleep(1)

        self._hb_register_count = 60
        # run heartbeat thread in all ranks (one hb msg every `n` seconds)
        self._log.debug('hb delay: %s', self._hb_delay)
        self._hb_thread = mt.Thread(target=self._hb_worker)
        self._hb_thread.daemon = True
        self._hb_thread.start()

        # run worker initialization *before* starting to work on requests.
        # the worker provides these builtin methods:
        #     eval:  evaluate a piece of python code with `eval`
        #     exec:  evaluate a piece of python code with `exec`
        #     call:  execute  a method or function call
        #     proc:  execute  a command line (fork/exec)
        #     shell: execute  a shell command
        self._modes = dict()
        self.register_mode(TASK_FUNC,  self._dispatch_func)
        self.register_mode(TASK_METH,  self._dispatch_meth)
        self.register_mode(TASK_EVAL,  self._dispatch_eval)
        self.register_mode(TASK_EXEC,  self._dispatch_exec)
        self.register_mode(TASK_PROC,  self._dispatch_proc)
        self.register_mode(TASK_SHELL, self._dispatch_shell)

        # prepare base env dict used for all tasks
        # NOTE: raptor tasks run in the same environment as the raptor worker
        self._task_env = dict()
        for k,v in os.environ.items():
            if not k.startswith('RP_'):
                self._task_env[k] = v

        reg_msg = {'cmd': 'worker_register',
                   'arg': {'uid'        : self._uid,
                           'raptor_id'  : self._raptor_id,
                           'ranks'      : self._ranks}}

        # the manager (rank 0) registers the worker with the master
        if self._manager:

            self._log.debug('register: %s / %s', self._uid, self._raptor_id)
            self._ctrl_pub.put(rpc.CONTROL_PUBSUB, reg_msg)

          # # FIXME: we never unregister on termination
          # self._ctrl_pub.put(rpc.CONTROL_PUBSUB, {'cmd': 'worker_unregister',
          #                                         'arg': {'uid' : self._uid}})

        # wait for raptor response (*all* ranks*)
        self._log.debug('wait for registration to complete')
        count = 0
        while not self._reg_event.wait(timeout=5):
            if count < self._hb_register_count:
                count += 1
                if self._manager:
                    self._log.debug('re-register: %s / %s', self._uid, self._raptor_id)
                    self._ctrl_pub.put(rpc.CONTROL_PUBSUB, reg_msg)
            else:
                self.stop()
                self.join()
                self._log.error('registration with master timed out')
                raise RuntimeError('registration with master timed out')

        if self._manager:
            self._log.debug('registration with master ok')


    # --------------------------------------------------------------------------
    #
    def _hb_worker(self):

        while True:

            self._ctrl_pub.put(rpc.CONTROL_PUBSUB,
                    {'cmd': 'worker_rank_heartbeat',
                     'arg': {'uid' : self._uid,
                             'rank': self._rank}})

            time.sleep(self._hb_delay)


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, topic, msgs):

        for msg in ru.as_list(msgs):

            cmd = msg['cmd']
            arg = msg['arg']

            if cmd != 'update':
                continue

            for thing in arg:

                uid   = thing['uid']
                state = thing['state']

                if uid == self._raptor_id:

                    if state in rps.FINAL + [rps.AGENT_STAGING_OUTPUT_PENDING]:
                        # master completed - terminate this worker
                        self._log.info('master %s final: %s - terminate',
                                       uid, state)
                        self.stop()
                        return False

        return True


    # --------------------------------------------------------------------------
    #
    def _control_cb(self, topic, msg):

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        if cmd == 'worker_registered':

            if arg['uid'] != self._uid:
                return

            if self._reg_event.is_set():
                # registration was completed already
                return

            self._ts_addr      = arg['info']['ts_addr']
            self._res_addr_put = arg['info']['res_addr_put']
            self._req_addr_get = arg['info']['req_addr_get']

            self._reg_event.set()

        elif cmd == 'terminate':
            self.stop()
            self.join()
            sys.exit()

        elif cmd == 'worker_terminate':

            if arg['uid'] == self._uid:
                self._log.debug('worker_terminate signal')
                self.stop()
                self.join()
                sys.exit()


    # --------------------------------------------------------------------------
    #
    def get_master(self):
        '''
        The worker can submit tasks back to the master - this method will
        return a small shim class to provide that capability.  That class has
        a single method `run_task` which accepts a single `rp.TaskDescription`
        from which a `rp.Task` is created and executed.  The call then waits for
        the task's completion before returning it in a dict representation, the
        same as when passed to the master's `result_cb`.

        Note: the `run_task` call is running in a separate thread and will thus
              not block the master's progress.

        Returns:
            Master: a shim class with only one method: `run_task(td)` where
                `td` is a `TaskDescription` to run.
        '''

        # ----------------------------------------------------------------------
        class Master(object):

            def __init__(self, addr):
                self._task_service_ep = ru.zmq.Client(url=addr)

            def run_task(self, td):
                return self._task_service_ep.request('run_task', td)
        # ----------------------------------------------------------------------

        return Master(self._ts_addr)


    # --------------------------------------------------------------------------
    #
    def start(self):
        '''Start the workers main work loop.
        '''

        raise NotImplementedError('`start()` must be implemented by child class')


    # --------------------------------------------------------------------------
    #
    def stop(self):
        '''Signal the workers to stop the main work loop.
        '''

        raise NotImplementedError('`stop()` must be implemented by child class')


    # --------------------------------------------------------------------------
    #
    def join(self):
        '''Wait until the worker's main work loop completed.
        '''

        raise NotImplementedError('`join()` must be implemented by child class')


    # --------------------------------------------------------------------------
    #
    def register_mode(self, name, dispatcher) -> None:
        '''
        Register a new task execution mode that this worker can handle.
        The specified dispatcher callable should accept a single argument: the
        task to execute.

        Args:
            name (str): name of the mode to register
            dispatcher (callable): function which implements the execution mode
        '''

        if name in self._modes:
            raise ValueError('mode %s already registered' % name)

        self._modes[name] = dispatcher


    # --------------------------------------------------------------------------
    #
    def get_dispatcher(self, name):
        '''Query a registered execution mode.

        Args:
            name (str): name of execution mode to query for

        Returns:
            Callable: the dispatcher method for that execution mode
        '''

        if name not in self._modes:
            raise ValueError('mode %s unknown' % name)

        return self._modes[name]


    # --------------------------------------------------------------------------
    #
    def _dispatch_meth(self, task):
        '''
        _dispatch_meth is a simple wrapper around _dispatch_func which points to
        private methods to be called.
        '''

        task['description']['function'] = task['description']['method']

        return self._dispatch_func(task)


    # --------------------------------------------------------------------------
    #
    def _dispatch_func(self, task):
        '''
        We expect three attributes: 'function', containing the name of the
        member method or free function to call, `args`, an optional list of
        unnamed parameters, and `kwargs`, and optional dictionary of named
        parameters.

        *function* is resolved first against `locals()`, then `globals()`, then
        attributes of the implementation class (member functions of *base*, as
        provided to `MPIWorkerRank()`). Finally, an attempt is made to
        deserialize a PythonTask from *function*. The first non-null resolution
        of *function* is used as the callable.

        NOTE: MPI function tasks will get a private communicator passed as first
              unnamed argument.

        Args:
            task (Dict[str, Any]): dictionary representation of the task to
                execute

        Returns:
            Tuple[str, str, int, Any, Tuple[str, str]]:
                - standard output (str)
                - standard error (str)
                - exit code (int)
                - return value (Any)
                - exception (Tuple[type (str), message (str)])

        Raises:
            KeyError
                if the task dictionary misses required entries
            ValueError
                if `task['description']['function']` cannot be resolved
            Assert
                if `task['description']['function']` is not set
        '''

        uid  = task['uid']
        func = task['description']['function']
        assert func

        args    = task['description'].get('args',   [])
        kwargs  = task['description'].get('kwargs', {})
        py_func = False

        self._log.debug('orig args: %s : %s', args, kwargs)

        # check if `func_name` is a global name
        names   = dict(list(globals().items()) + list(locals().items()))
        to_call = names.get(func)

        # if not, check if this is a class method of this worker implementation
        if not to_call:
            to_call = getattr(self, func, None)

        # check if we have a serialized object
        if not to_call:
            self._log.debug('func serialized: %d: %s', len(func), func)

            try:
                to_call, _args, _kwargs = PythonTask.get_func_attr(func)

            except Exception:
                self._log.warn('function is not a PythonTask [%s] ', uid)

            else:
                py_func = True
                if args or kwargs:
                    raise ValueError('`args` and `kwargs` must be empty for'
                                     'PythonTask function [%s]' % uid)
                else:
                    args   = _args
                    kwargs = _kwargs

        if not to_call:
            self._log.error('no %s in \n%s\n\n%s', func, names, dir(self))
            raise ValueError('%s callable %s not found: %s' % (uid, func, task))

        comm = task.get('mpi_comm')
        if comm:
            # we have an MPI communicator we need to inject into the function's
            # arguments.
            if py_func:
                # For a `py_func` we add the communicator as `comm` kwarg if
                # that is set to None, and otherwise as first `arg` if that is
                # None.  If neither is true we'll error out.
                # NOTE that we don't change the number of arguments either way.
                if 'comm' in kwargs and kwargs['comm'] is None:
                    kwargs['comm'] = comm
                elif args and args[0] is None:
                    args[0] = comm
                else:
                    raise RuntimeError("can't inject communicator for %s: %s: %s",
                                       task['uid'], args, kwargs)
            else:
                args.insert(0, comm)

        # make sure we capture stdout / stderr
        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        # set the task environment
        old_env = os.environ.copy()

        for k, v in task['description'].get('environment', {}).items():
            os.environ[k] = str(v)

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            self._prof.prof('rank_start', uid=uid)
            self._log.debug('to call %s: %s : %s', to_call, args, kwargs)
            val = to_call(*args, **kwargs)
            self._prof.prof('rank_stop', uid=uid)
            out = strout.getvalue()
            err = strerr.getvalue()
            exc = (None, None)
            ret = 0

        except Exception as e:
            self._log.exception('_call failed: %s', task['uid'])
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\ncall failed: %s' % e)
            exc = (repr(e), '\n'.join(ru.get_exception_trace()))
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr

            # remove communicator from args again
            if comm:
                if py_func:
                    if 'comm' in kwargs:
                        del kwargs['comm']
                    elif args:
                        args[0] = None
                else:
                    args.pop(0)

            os.environ = old_env

        self._log.debug('%s: got %s', uid, out)

        return out, err, ret, val, exc


    # --------------------------------------------------------------------------
    #
    def _dispatch_eval(self, task):
        '''
        We expect a single attribute: 'code', containing the Python
        code to be eval'ed
        '''

        uid  = task['uid']
        code = task['description']['code']
        assert code

        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        old_env = os.environ.copy()

        for k, v in task['description'].get('environment', {}).items():
            os.environ[k] = str(v)

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            self._log.debug('eval [%s] [%s]', code, task['uid'])

            self._prof.prof('rank_start', uid=uid)
            val = eval(code)
            self._prof.prof('rank_stop', uid=uid)
            out = strout.getvalue()
            err = strerr.getvalue()
            exc = (None, None)
            ret = 0

        except Exception as e:
            self._log.exception('_eval failed: %s', task['uid'])
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\neval failed: %s' % e)
            exc = (repr(e), '\n'.join(ru.get_exception_trace()))
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr

            os.environ = old_env

        return out, err, ret, val, exc



    # --------------------------------------------------------------------------
    #
    def _dispatch_exec(self, task):
        '''
        We expect a single attribute: 'code', containing the Python code to be
        exec'ed.  The optional attribute `pre_exec` can be used for any import
        statements and the like which need to run before the executed code.
        '''

        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        old_env = os.environ.copy()

        for k, v in task['description'].get('environment', {}).items():
            os.environ[k] = str(v)

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            uid  = task['uid']
            pre  = task['description'].get('pre_exec', [])
            code = task['description']['code']

            # create a wrapper function around the given code
            lines = code.split('\n')
            outer = 'def _my_exec():\n'
            for line in lines:
                outer += '    ' + line + '\n'

            # call that wrapper function via exec, and keep the return value
            src = '%s\n\n%s\n\nresult=_my_exec()' % ('\n'.join(pre), outer)

            # assign a local variable to capture the code's return value.
            loc = dict()
            self._prof.prof('rank_start', uid=uid)
            exec(src, {}, loc)                # pylint: disable=exec-used # noqa
            self._prof.prof('rank_stop', uid=uid)
            val = loc['result']
            out = strout.getvalue()
            err = strerr.getvalue()
            exc = (None, None)
            ret = 0

        except Exception as e:
            self._log.exception('_exec failed: %s', task['uid'])
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\nexec failed: %s' % e)
            exc = (repr(e), '\n'.join(ru.get_exception_trace()))
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr

            os.environ = old_env

        return out, err, ret, val, exc


    # --------------------------------------------------------------------------
    #
    def _dispatch_proc(self, task):
        '''
        We expect two attributes: 'executable', containing the executabele to
        run, and `arguments` containing a list of arguments (strings) to pass as
        command line arguments.  We use `sp.Popen` to run the fork/exec, and to
        collect stdout, stderr and return code
        '''

        try:
            import subprocess as sp

            uid  = task['uid']
            exe  = task['description']['executable']
            args = task['description'].get('arguments', list())
            env  = dict(self._task_env)
            env.update(task['description']['environment'])

            cmd  = '%s %s' % (exe, ' '.join([shlex.quote(arg) for arg in args]))
            self._prof.prof('rank_start', uid=uid)
            proc = sp.Popen(cmd, env=env,  stdin=None,
                            stdout=sp.PIPE, stderr=sp.PIPE,
                            close_fds=True, shell=True)
            out, err = proc.communicate()
            ret      = proc.returncode
            exc      = (None, None)
            self._prof.prof('rank_stop', uid=uid)

        except Exception as e:
            self._log.exception('proc failed: %s', task['uid'])
            out = None
            err = 'exec failed: %s' % e
            exc = (repr(e), '\n'.join(ru.get_exception_trace()))
            ret = 1

        return out, err, ret, None, exc


    # --------------------------------------------------------------------------
    #
    def _dispatch_shell(self, task):
        '''
        We expect a single attribute: 'command', containing the command
        line to be called as string.
        '''

        try:
            uid = task['uid']
            cmd = task['description']['command']
            env = dict(self._task_env)
            env.update(task['description']['environment'])

          # self._log.debug('shell: --%s--', cmd)

            self._prof.prof('rank_start', uid=uid)
            out, err, ret = ru.sh_callout(cmd, shell=True, env=env)
            exc = (None, None)
            self._prof.prof('rank_stop', uid=uid)

        except Exception as e:
            self._log.exception('_shell failed: %s', task['uid'])
            out = None
            err = 'shell failed: %s' % e
            exc = (repr(e), '\n'.join(ru.get_exception_trace()))
            ret = 1

      # os.environ = old_env

        return out, err, ret, None, exc


    # --------------------------------------------------------------------------
    #
    def hello(self, msg, sleep=0):

        print('hello %s: %.3f' % (msg, time.time()))
        time.sleep(sleep)
        print('hello %s: %.3f' % (msg, time.time()))
        return 'hello %s' % msg


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import copy
import time

from typing import Optional

import threading        as mt

import radical.utils    as ru

from . import constants as rpc
from . import utils     as rpu

from .messages        import HeartbeatMessage
from .proxy           import Proxy
from .resource_config import ResourceConfig, ENDPOINTS_DEFAULT


# ------------------------------------------------------------------------------
#
class _CloseOptions(ru.TypedDict):
    """Options and validation for Session.close().

    Arguments:
        download (bool, optional): Fetch pilot profiles and database entries.
            (Default False.)
        terminate (bool, optional): Shut down all pilots associated with the
            session. (Default True.)

    """

    _check = True

    _schema = {
        'download' : bool,
        'terminate': bool,
        'cleanup'  : bool  # FIXME: to be removed
    }

    _defaults = {
        'download' : False,
        'terminate': True,
        'cleanup'  : True  # FIXME: to be removed
    }


# ------------------------------------------------------------------------------
#
class Session(object):
    """Root of RP object hierarchy for an application instance.

    A Session is the root object of all RP objects in an application instance:
    it holds :class:`radical.pilot.PilotManager` and
    :class:`radical.pilot.TaskManager` instances which in turn hold
    :class:`radical.pilot.Pilot` and :class:`radical.pilot.Task`
    instances, and several other components which operate on those stateful
    entities.
    """

    # In that role, the session will create a special pubsub channel `heartbeat`
    # which is used by all components in its hierarchy to exchange heartbeat
    # messages.  Those messages are used to watch component health - if
    # a (parent or child) component fails to send heartbeats for a certain
    # amount of time, it is considered dead and the process tree will terminate.
    # That heartbeat management is implemented in the `ru.Heartbeat` class.
    # Only primary sessions instantiate a heartbeat channel (i.e., only the root
    # sessions of RP client or agent modules), but all components need to call
    # the sessions `heartbeat()` method at regular intervals.

    # the reporter is an application-level singleton
    _reporter = None

    # a session has one of three possible roles:
    #   - primary: the session is the first explicit session instance created in
    #     an RP application.
    #   - agent: the session is the first session instance created in an RP
    #     agent.
    #   - default: any other session instance, for example such as created by
    #     components in the client or agent module.
    _PRIMARY = 'primary'
    _AGENT_0 = 'agent_0'
    _AGENT_N = 'agent_n'
    _DEFAULT = 'default'


    # --------------------------------------------------------------------------
    #
    def __init__(self, proxy_url: Optional[str ] = None,
                       uid      : Optional[str ] = None,
                       cfg      : Optional[dict] = None,
                       _role    : Optional[str ] = _PRIMARY,
                       _reg_addr: Optional[str ] = None,
                       **close_options):
        """Create a new session.

        A new Session instance is created and stored in the database.

        Any RP Session will require an RP Proxy to facilitate communication
        between the client machine (i.e., the host where the application created
        this Session instance) and the target resource (i.e., the host where the
        pilot agent/s is/are running and where the workload is being executed).

        A `proxy_url` can be specified which then must point to an RP Proxy
        Service instance which this session can use to establish a communication
        proxy.  If `proxy_url` is not specified, the session will check for the
        environment variables `RADICAL_PILOT_PROXY_URL` and will interpret it as
        such above.  If that information is not available, the session will
        instantiate a proxy service on the local host.  Note that any proxy
        service instantiated by the session itself will be terminated once the
        session instance is closed or goes out of scope and is thus garbage
        collected and as such should not be used by other session instances.

        Note: an RP proxy will have to be accessible by both the client and the
              target hosts to facilitate communication between both parties.
              That implies access to the respective ports.  Proxies started by
              the session itself will use the first port larger than 10.000
              which is found to be free.

        Arguments:

            proxy_url (str, optional): proxy service URL - points to an RP
              proxy service which is used to establish an RP communication proxy
              for this session.

            uid (str, optional): Create a session with this UID.  Session UIDs
                MUST be unique - otherwise they will lead to communication
                conflicts, resulting in undefined behaviours.

            cfg (str | dict, optional): a named or instantiated configuration
                to be used for the session.

            _role (`bool`): only `PRIMARY` sessions created by the original
                application process (via `rp.Session()`), will create proxies
                and Registry Services.  `AGENT` sessions will also create
                a Registry but no proxies.  All other `DEFAULT` session
                instances are instantiated internally in processes spawned
                (directly or indirectly) by the initial session, for example in
                some of it's components, or by the RP agent.  Those sessions
                will inherit the original session ID, but will not attempt to
                create a new proxies or registries.

            **close_options (optional): If additional key word arguments are
                provided, they will be used as the default arguments to
                Session.close(). This can be useful when the Session is used as
                a Python context manager, such that close() is called
                automatically at the end of a ``with`` block.

            _reg_addr (str, optional): Non-primary sessions will connect to the
                registry at that endpoint and pull session config and resource
                configurations from there.
        """
        self._t_start = time.time()

        self._role          = _role
        self._uid           = uid
        self._cfg           = ru.Config(cfg=cfg)
        self._reg_addr      = _reg_addr
        self._proxy_url     = proxy_url
        self._proxy_cfg     = None
        self._closed        = False
        self._created       = time.time()
        self._close_options = _CloseOptions(close_options)
        self._close_options.verify()

        self._proxy    = None    # proxy client instance
        self._reg      = None    # registry client instance
        self._pmgrs    = dict()  # map IDs to pmgr instances
        self._tmgrs    = dict()  # map IDs to tmgr instances
        self._cmgr     = None    # only primary sessions have a cmgr
        self._rm       = None    # resource manager (agent_0 sessions)
        self._hb       = None    # heartbeat monitor

        # this session is either living in the client application or lives in
        # the scope of a pilot.  In the latter case we expect `RP_PILOT_ID` to
        # be set - we derive the session module scope from that env variable.
        self._module = os.environ.get('RP_PILOT_ID', 'client')

        # non-primary sessions need a uid!
        if self._role != self._PRIMARY and not self._uid:
            raise ValueError('non-primary session needs UID (%s)' % self._role)

        # initialization is different for each session type
        # NOTE: we could refactor this to session sub-classes
        if   self._role == self._PRIMARY: self._init_primary()
        elif self._role == self._AGENT_0: self._init_agent_0()
        elif self._role == self._AGENT_N: self._init_agent_n()
        else                            : self._init_default()

        # cache sandboxes etc.
        self._cache_lock = ru.RLock()
        self._cache      = {'endpoint_fs'      : dict(),
                            'resource_sandbox' : dict(),
                            'session_sandbox'  : dict(),
                            'pilot_sandbox'    : dict(),
                            'client_sandbox'   : self._cfg.client_sandbox,
                            'js_shells'        : dict(),
                            'fs_dirs'          : dict()}

        # at this point we have a bridge connection, logger, etc, and are done
        self._prof.prof('session_ok', uid=self._uid)

        if self._role == self._PRIMARY:
            self._rep.ok('>>ok\n')

        assert(self._reg)


    # --------------------------------------------------------------------------
    #
    def _init_primary(self):

        assert self._role == self._PRIMARY

        # The primary session
        #   - reads session config files
        #   - reads resource config files
        #   - starts the client side registry service
        #   - pushes the configs into that registry
        #   - pushes bridge and component configs into that registry
        #   - starts a ZMQ proxy (or ensures one is up and running)

        # if user did not set a uid, we need to generate a new ID
        if not self._uid:
            self._uid = ru.generate_id('rp.session', mode=ru.ID_PRIVATE)

        # we still call `_init_cfg` to complete missing config settings
        # FIXME: completion only needed by `PRIMARY`
        self._init_cfg_from_scratch()

        # primary sessions create a registry service
        self._start_registry()
        self._connect_registry()

        # only primary sessions start and initialize the proxy service
        self._start_proxy()

        # start heartbeat channel
        self._start_heartbeat()

        # push the session config into the registry
        self._publish_cfg()

        # start bridges and components
        self._start_components()

        time.sleep(1)

        # primary session hooks into the control pubsub
        bcfg = self._reg['bridges.%s' % rpc.CONTROL_PUBSUB]
        self._ctrl_pub = ru.zmq.Publisher(channel=rpc.CONTROL_PUBSUB,
                                          url=bcfg['addr_pub'],
                                          log=self._log,
                                          prof=self._prof)

        # crosswire local channels and proxy channels
        self._crosswire_proxy()


    # --------------------------------------------------------------------------
    #
    def _init_agent_0(self):

        # The agent_0 session expects the `cfg` parameter to contain the
        # complete agent config!
        #
        #   - starts the agent side registry service
        #   - separates
        #     - session config (== agent config)
        #     - bridge configs
        #     - component configs
        #     - resource config
        #   - pushes them all into the registry
        #   - connects to the ZMQ proxy for client/agent communication
        #   - start agent components

        assert self._role == self._AGENT_0

        self._init_cfg_from_dict()
        self._start_registry()
        self._connect_registry()
        self._connect_proxy()
        self._start_heartbeat()
        self._publish_cfg()
        self._init_rm()
        self._start_components()
        self._crosswire_proxy()


    # --------------------------------------------------------------------------
    #
    def _init_agent_n(self):

        # The agent_n session fetch their config from agent_0 registry
        #
        #   - connect to registry
        #   - fetch config from registry
        #   - start agent bridges and components

        assert self._role == self._AGENT_N

        # the config passed to the session c'tor is the *agent* config - keep it
        a_cfg = self._cfg

        self._connect_registry()
        self._init_cfg_from_registry()

        # merge the agent's config into the session config
        self._cfg.bridges    = ru.Config(cfg=a_cfg.get('bridges',    {}))
        self._cfg.components = ru.Config(cfg=a_cfg.get('components', {}))

        self._start_components()


    # --------------------------------------------------------------------------
    #
    def _init_default(self):

        # sub-agents and components connect to an existing registry (owned by
        # the `primary` session or `agent_0`) and load config settings from
        # there.

        assert self._role == self._DEFAULT

        self._connect_registry()
        self._init_cfg_from_registry()


    # --------------------------------------------------------------------------
    #
    def _start_registry(self):

        # make sure that no other registry is used
        if self._reg_addr:
            raise ValueError('cannot start registry when providing `reg_addr`')

        self._reg_service = ru.zmq.Registry(uid='%s.reg' % self._uid,
                                            path=self._cfg.path)
        self._reg_service.start()

        self._cfg.reg_addr = self._reg_service.addr


    # --------------------------------------------------------------------------
    #
    def _connect_registry(self):

        if not self._cfg.reg_addr:
            self._cfg.reg_addr = self._reg_addr

        if not self._cfg.reg_addr:
            raise ValueError('session needs a registry address')

        self._reg = ru.zmq.RegistryClient(url=self._cfg.reg_addr)


    # --------------------------------------------------------------------------
    #
    def _init_cfg_from_scratch(self):

        # A primary session will at this point have a registry client connected
        # to its registry service.  Further, self._cfg will either be a config
        # name to be read from disk (`session_<cfg_name>.json`), or a dictionary
        # with a specific, user provided config.  From this information clean up
        # `self._cfg` and store it in the registry.  Also read resource configs
        # and store the in the registry as well.

        # NOTE: `cfg_name` and `cfg` are overloaded, the user cannot point to
        #       a predefined config and amend it at the same time.  This might
        #       be ok for the session, but introduces an API inconsistency.

        cfg_name = 'default'
        if isinstance(self._cfg, str):
            cfg_name  = self._cfg
            self._cfg = None

        # load the named config, merge provided config
        self._cfg = ru.Config('radical.pilot.session', name=cfg_name,
                                                       cfg=self._cfg)

        rcfgs = ru.Config('radical.pilot.resource', name='*', expand=False)
        self._rcfgs = ru.Config()

        for site in rcfgs:
            self._rcfgs[site] = ru.Config()
            for res, rcfg in rcfgs[site].items():
                self._rcfgs[site][res] = ResourceConfig(rcfg)


        self._rcfg  = ru.Config()  # the local resource config, if known

        # set essential config values for *this* specific session
        self._cfg['sid'] = self._uid

        pwd = os.getcwd()

        if self._cfg.base:
            self._cfg.base = os.path.abspath(self._cfg.base)
        else:
            self._cfg.base = pwd

        if self._cfg.path:
            self._cfg.path = os.path.abspath(self._cfg.path)
        else:
            self._cfg.path = '%s/%s' % (self._cfg.base, self._cfg.sid)

        if self._cfg.client_sandbox:
            self._cfg.client_sandbox = os.path.abspath(self._cfg.client_sandbox)
        else:
            self._cfg.client_sandbox = pwd

        # change RU defaults to point logfiles etc. to the session sandbox
        def_cfg             = ru.DefaultConfig()
        def_cfg.log_dir     = self._cfg.path
        def_cfg.report_dir  = self._cfg.path
        def_cfg.profile_dir = self._cfg.path

        self._prof = self._get_profiler(name=self._uid)
        self._rep  = self._get_reporter(name=self._uid)
        self._log  = self._get_logger  (name=self._uid,
                                        level=self._cfg.get('log_lvl'),
                                        debug=self._cfg.get('debug_lvl'))

        from . import version_detail as rp_version_detail
        self._log.info('radical.pilot version: %s', rp_version_detail)
        self._log.info('radical.utils version: %s', ru.version_detail)

        self._prof.prof('session_start', uid=self._uid)

        self._rep.info ('<<new session: ')
        self._rep.plain('[%s]' % self._uid)


    # --------------------------------------------------------------------------
    #
    def _init_cfg_from_dict(self):

        # A agent_0 session will read the configuration from agent_0.cfg and
        # pass it to the session.

        assert self._role == self._AGENT_0

        self._cfg = ru.Config(cfg=self._cfg)

        # we only have one resource config for the current resource
        self._rcfg  = ru.Config(cfg=self._cfg.resource_cfg)  # local config
        self._rcfgs = ru.Config()

        del self._cfg['resource_cfg']

        # set essential config values for *this* specific session
        pwd = os.getcwd()

        if self._cfg.base:
            self._cfg.base = os.path.abspath(self._cfg.base)
        else:
            self._cfg.base = pwd

        if self._cfg.path:
            self._cfg.path = os.path.abspath(self._cfg.path)
        else:
            self._cfg.path = pwd

        # change RU defaults to point logfiles etc. to the session sandbox
        def_cfg             = ru.DefaultConfig()
        def_cfg.log_dir     = self._cfg.path
        def_cfg.report_dir  = self._cfg.path
        def_cfg.profile_dir = self._cfg.path

        self._prof = self._get_profiler(name=self._uid)
        self._rep  = self._get_reporter(name=self._uid)
        self._log  = self._get_logger  (name=self._uid,
                                        level=self._cfg.get('log_lvl'),
                                        debug=self._cfg.get('debug_lvl'))

        from . import version_detail as rp_version_detail
        self._log.info('radical.pilot version: %s', rp_version_detail)
        self._log.info('radical.utils version: %s', ru.version_detail)

        self._prof.prof('session_start', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _init_cfg_from_registry(self):

        # fetch config settings from the registry
        self._cfg   = ru.Config(cfg=self._reg['cfg'])
        self._rcfg  = ru.Config(cfg=self._reg['rcfg'])
        self._rcfgs = ru.Config(cfg=self._reg['rcfgs'])

        # change RU defaults to point logfiles etc. to the session sandbox
        # NOTE: this is racey: the first session in this process will win
        def_cfg             = ru.DefaultConfig()
        def_cfg.log_dir     = self._cfg.path
        def_cfg.report_dir  = self._cfg.path
        def_cfg.profile_dir = self._cfg.path

        self._prof = self._get_profiler(name=self._uid)
        self._rep  = self._get_reporter(name=self._uid)
        self._log  = self._get_logger  (name=self._uid,
                                        level=self._cfg.get('log_lvl'),
                                        debug=self._cfg.get('debug_lvl'))

        from . import version_detail as rp_version_detail
        self._log.info('radical.pilot version: %s', rp_version_detail)
        self._log.info('radical.utils version: %s', ru.version_detail)

        self._log.debug('Session(%s, %s)', self._uid, self._role)
        self._prof.prof('session_start', uid=self._uid)


    # --------------------------------------------------------------------------
    #
    def _start_heartbeat(self):

        # only primary and agent_0 sessions manage heartbeats
        assert self._role in [self._PRIMARY, self._AGENT_0]

        # start the embedded heartbeat pubsub bridge
        self._hb_pubsub = ru.zmq.PubSub('heartbeat_pubsub',
                                        cfg={'uid'    : 'heartbeat_pubsub',
                                             'type'   : 'pubsub',
                                             'log_lvl': 'debug',
                                             'path'   : self._cfg.path})
        self._hb_pubsub.start()
        time.sleep(1)

        # re-enable the test below if timing issues crop up
      # ru.zmq.test_pubsub(self._hb_pubsub.channel,
      #                    self._hb_pubsub.addr_pub,
      #                    self._hb_pubsub.addr_sub),

        # fill 'cfg.heartbeat' section
        self._cfg.heartbeat.addr_pub = str(self._hb_pubsub.addr_pub)
        self._cfg.heartbeat.addr_sub = str(self._hb_pubsub.addr_sub)

        # create a publisher for that channel to publish own heartbeat
        self._hb_pub = ru.zmq.Publisher(channel='heartbeat_pubsub',
                                        url=self._cfg.heartbeat.addr_pub,
                                        log=self._log,
                                        prof=self._prof)


        # --------------------------------------
        # start the heartbeat monitor, but first
        # define its callbacks
        def _hb_beat_cb():
            # called on every heartbeat: cfg.heartbeat.interval`
            # publish own heartbeat
            self._hb_pub.put('heartbeat', HeartbeatMessage(uid=self._uid))

            # also update proxy heartbeat
            if self._proxy:
                self._proxy.request('heartbeat', {'sid': self._uid})
        # --------------------------------------

        # --------------------------------------
        # called when some entity misses
        # heartbeats: `cfg.heartbeat.timeout`
        def _hb_term_cb(hb_uid):
            if self._cmgr:
                self._cmgr.close()
            return False
        # --------------------------------------

        # create heartbeat manager which monitors all components in this session
      # self._log.debug('hb %s from session', self._uid)
        self._hb = ru.Heartbeat(uid=self._uid,
                                timeout=self._cfg.heartbeat.timeout,
                                interval=self._cfg.heartbeat.interval,
                                beat_cb=_hb_beat_cb,
                                term_cb=_hb_term_cb,
                                log=self._log)
        self._hb.start()

        # --------------------------------------
        # subscribe to heartbeat msgs and inform
        # self._hb about every heartbeat
        def _hb_msg_cb(topic, msg):

            hb_msg = HeartbeatMessage(from_dict=msg)

            if hb_msg.uid != self._uid:
                self._hb.beat(uid=hb_msg.uid)
        # --------------------------------------

        ru.zmq.Subscriber(channel='heartbeat_pubsub',
                          topic='heartbeat',
                          url=self._cfg.heartbeat.addr_sub,
                          cb=_hb_msg_cb,
                          log=self._log,
                          prof=self._prof)


    # --------------------------------------------------------------------------
    #
    def _publish_cfg(self):

        # The primary session and agent_0 push their configs into the registry.
        # NOTE: proxy channels populate the registrie's `bridges` section

        assert self._role in [self._PRIMARY, self._AGENT_0]

        # push proxy, bridges, components and heartbeat subsections separately
        flat_cfg = copy.deepcopy(self._cfg)

        del flat_cfg['heartbeat']
        del flat_cfg['bridges']
        del flat_cfg['components']

        self._reg['cfg']        = flat_cfg
        self._reg['heartbeat']  = self._cfg.heartbeat
        self._reg['bridges']    = self._cfg.bridges  # proxy bridges
        self._reg['components'] = {}

        # if we have proxy channels, publish them in the bridges configs too
        if self._proxy_cfg:
            for channel in self._proxy_cfg:
                self._reg['bridges.%s' % channel] = self._proxy_cfg[channel]

        # primary sessions publish all known resource configs under `rcfgs`, the
        # agent_0 only publishes the *current* resource config under `rcfg`.
        if self._role == self._PRIMARY:
            self._reg['rcfg']  = dict()
            self._reg['rcfgs'] = self._rcfgs

        elif self._role == self._AGENT_0:
            self._reg['rcfg']  = self._rcfg
            self._reg['rcfgs'] = dict()


    # --------------------------------------------------------------------------
    #
    def _start_proxy(self):

        # A primary session will start a ZMQ proxy via which agents can connect
        # to the client.  It is also possible a proxy already exists and a proxy
        # address was passed to the ctor - in that case we will, obviously, not
        # start a proxy but just make sure the given one is being used.

        assert self._role == self._PRIMARY

        # check if an external proxy was epcified for the session.
        if self._proxy_url:
            self._log.debug('use proxy at %s' % self._proxy_url)
            return

       # check if an external proxy was specified in the environment
        elif 'RADICAL_PILOT_PROXY_URL' in os.environ:
            self._proxy_url = os.environ['RADICAL_PILOT_PROXY_URL']
            self._log.debug('found proxy at %s' % self._proxy_url)

        # no luck - start an embedded proxy
        else:
            self._proxy_event  = mt.Event()
            self._proxy_thread = mt.Thread(target=self._run_proxy)
            self._proxy_thread.daemon = True
            self._proxy_thread.start()

            self._proxy_event.wait()


        assert self._proxy_url

        # the proxy url becomes part of the session cfg
        self._cfg.proxy_url = self._proxy_url

        self._rep.info ('<<zmq proxy  : ')
        self._rep.plain('[%s]' % self._proxy_url)

        # configure proxy channels
        try:
            self._proxy = ru.zmq.Client(url=self._cfg.proxy_url)
            self._proxy_cfg = self._proxy.request('register', {'sid':self._uid})

        except:
            self._log.exception('%s: failed to start proxy', self._role)
            raise


    # --------------------------------------------------------------------------
    #
    def _connect_proxy(self):

        assert self._role == self._AGENT_0

        # make sure we have a proxy address to use
        assert self._cfg.proxy_url

        # query the proxy service to fetch proxy cfg created by primary session
        self._proxy = ru.zmq.Client(url=self._cfg.proxy_url)
        self._proxy_cfg = self._proxy.request('lookup', {'sid': self._uid})
        self._log.debug('proxy response: %s', self._proxy_cfg)


    # ----------------------------------------------------------------------
    def crosswire_pubsub(self, src, tgt, from_proxy):

        # we only forward messages which have either no origin set (in this case
        # this method sets the origin), or whose origin is the same as
        # configured when crosswiring the channels (either 'client' or the pilot
        # ID).  Also, the messages need to have the `forward` flag set.

        path = self._cfg.path
        reg  = self._reg

        url_sub = reg['bridges.%s.addr_sub' % src.lower()]
        url_pub = reg['bridges.%s.addr_pub' % tgt.lower()]

      # self._log.debug('XXX cfg fwd for topic:%s to %s', src, tgt)
      # self._log.debug('XXX cfg fwd for %s to %s', url_sub, url_pub)

        publisher = ru.zmq.Publisher(channel=tgt, path=path, url=url_pub,
                                     log=self._log, prof=self._prof)

        def pubsub_fwd(topic, msg):

            if 'origin' not in msg:
                msg['origin'] = self._module

            if from_proxy:

                # all messages *from* the proxy are forwarded - but not the ones
                # which originated in *this* module in the first place.

                if msg['origin'] == self._module:
                  # self._log.debug('XXX >=! fwd %s to topic:%s: %s', src, tgt, msg)
                    return

              # self._log.debug('XXX >=> fwd %s to topic:%s: %s', src, tgt, msg)
                publisher.put(tgt, msg)

            else:

                # only forward messages which have the respective flag set
                if not msg.get('fwd'):
                  # self._log.debug('XXX =>! fwd %s to %s: %s [%s - %s]', src,
                  #                 tgt, msg, msg['origin'], self._module)
                    return

                # only forward all messages which originated in *this* module.
                if not msg['origin'] == self._module:
                  # self._log.debug('XXX =>| fwd %s to topic:%s: %s', src, tgt, msg)
                    return

                # avoid message loops (forward only once)
                msg['fwd'] = False

              # self._log.debug('XXX =>> fwd %s to topic:%s: %s', src, tgt, msg)
                publisher.put(tgt, msg)


        ru.zmq.Subscriber(channel=src, topic=src, path=path, cb=pubsub_fwd,
                          url=url_sub, log=self._log, prof=self._prof)


    # --------------------------------------------------------------------------
    #
    def _crosswire_proxy(self):

        # - forward local ctrl  pubsub messages to proxy control pubsub
        # - forward local state pubsub messages to proxy state   pubsub
        # - forward local task  queue  messages to proxy task    queue
        #
        # - forward proxy ctrl  pubsub messages to local control pubsub
        # - forward proxy state pubsub messages to local state   pubsub
        # - forward proxy task  queue  messages to local task    queue
        #
        # The local task queue endpoints differ for primary session and agent_0
        #
        # - primary:
        #   - forward from AGENT_STAGING_INPUT_PENDING_QUEUE
        #   - forward to   TMGR_STAGING_OUTPUT_PENDING_QUEUE
        #
        # - agent_0:
        #   - forward to   AGENT_STAGING_INPUT_PENDING_QUEUE
        #   - forward from TMGR_STAGING_OUTPUT_PENDING_QUEUE
        #
        # NOTE: the primary session task queues don't live in the session itself
        #       but are owned by the task manager instead - it will trigger the
        #       crosswire once the queues are created.

        assert self._role in [self._PRIMARY, self._AGENT_0]

        self.crosswire_pubsub(src=rpc.CONTROL_PUBSUB,
                              tgt=rpc.PROXY_CONTROL_PUBSUB,
                              from_proxy=False)
        self.crosswire_pubsub(src=rpc.PROXY_CONTROL_PUBSUB,
                              tgt=rpc.CONTROL_PUBSUB,
                              from_proxy=True)

        self.crosswire_pubsub(src=rpc.STATE_PUBSUB,
                              tgt=rpc.PROXY_STATE_PUBSUB,
                              from_proxy=False)
        self.crosswire_pubsub(src=rpc.PROXY_STATE_PUBSUB,
                              tgt=rpc.STATE_PUBSUB,
                              from_proxy=True)


    # --------------------------------------------------------------------------
    #
    def _init_rm(self):

        # import locally to avoid circular imports
        from .agent.resource_manager import ResourceManager

        rname    = self._rcfg.resource_manager
        self._rm = ResourceManager.create(name=rname,
                                          cfg=self._cfg,
                                          rcfg=self._rcfg,
                                          log=self._log, prof=self._prof)

        import pprint
        self._log.debug(pprint.pformat(self._rm.info))


    # --------------------------------------------------------------------------
    #
    def get_rm(self):

        return self._rm


    # --------------------------------------------------------------------------
    #
    def _start_components(self):

        assert self._role in [self._PRIMARY, self._AGENT_0, self._AGENT_N]

        # primary sessions and agents have a component manager which also
        # manages heartbeat.  'self._cmgr.close()` should be called during
        # termination
        self._cmgr = rpu.ComponentManager(self.uid, self.reg_addr, self._uid)
        self._cmgr.start_bridges(self._cfg.bridges)
        self._cmgr.start_components(self._cfg.components)


    # --------------------------------------------------------------------------
    # context manager `with` clause
    #
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


    # --------------------------------------------------------------------------
    #
    def close(self, **kwargs):
        """Close the session.

        All subsequent attempts access objects attached to the session will
        result in an error. If cleanup is set to True, the session data is
        removed from the database.

        Arguments:
            terminate (bool, optional): Shut down all pilots associated with the
                session.
            download (bool, optional): Fetch pilot profiles and database
                entries.
        """

        # close only once
        if self._closed:
            return

        if self._role == self._PRIMARY:
            self._rep.info('closing session %s' % self._uid)

        self._log.debug("session %s closing", self._uid)
        self._prof.prof("session_close", uid=self._uid)

        # Merge kwargs with current defaults stored in self._close_options
        self._close_options.update(kwargs)
        self._close_options.verify()

        # to call for `_verify` method and to convert attributes
        # to their types if needed (but None value will stay if it is set)

        options = self._close_options

        if options.terminate:
            # terminate all components
            if self._role == self._PRIMARY:
                self._ctrl_pub.put(rpc.CONTROL_PUBSUB, {'cmd': 'terminate',
                                                        'arg': None})

        for tmgr_uid, tmgr in self._tmgrs.items():
            self._log.debug("session %s closes tmgr   %s", self._uid, tmgr_uid)
            tmgr.close()
            self._log.debug("session %s closed tmgr   %s", self._uid, tmgr_uid)

        for pmgr_uid, pmgr in self._pmgrs.items():
            self._log.debug("session %s closes pmgr   %s", self._uid, pmgr_uid)
            pmgr.close(terminate=options.terminate)
            self._log.debug("session %s closed pmgr   %s", self._uid, pmgr_uid)

        if self._cmgr:
            self._cmgr.close()

        # stop heartbeats
        if self._hb:
            self._hb.stop()
            self._hb_pubsub.stop()

        if self._proxy:

            if self._role == self._PRIMARY:
                try:
                    self._log.debug('session %s closes service', self._uid)
                    self._proxy.request('unregister', {'sid': self._uid})
                except:
                    pass

            if self._role in [self._PRIMARY, self._AGENT_0]:
                self._proxy.close()
                self._proxy = None

        self._log.debug("session %s closed", self._uid)
        self._prof.prof("session_stop", uid=self._uid)
        self._prof.close()

        self._closed = True

        # after all is said and done, we attempt to download the pilot log- and
        # profiles, if so wanted
        if options.download:

            self._prof.prof("session_fetch_start", uid=self._uid)
            self._log.debug('start download')
            tgt = self._cfg.base
            self.fetch_profiles(tgt=tgt)
            self.fetch_logfiles(tgt=tgt)

            self._prof.prof("session_fetch_stop", uid=self._uid)

        if self._role == self._PRIMARY:

            # stop registry
            self._reg.close()
            self._reg_service.stop()  # this will dump registry

            self._t_stop = time.time()
            self._rep.info('<<session lifetime: %.1fs'
                          % (self._t_stop - self._t_start))
            self._rep.ok('>>ok\n')


    # --------------------------------------------------------------------------
    #
    def _run_proxy(self):

        proxy = Proxy(path=self._cfg.path)

        try:
            proxy.start()

            self._proxy_url = proxy.addr
            self._proxy_event.set()

            # run forever until process is interrupted or killed
            proxy.wait()

        finally:
            proxy.stop()
            proxy.wait()


    # --------------------------------------------------------------------------
    #
    def as_dict(self):
        """Returns a Python dictionary representation of the object."""

        object_dict = {
            'uid'      : self._uid,
            'proxy_url': str(self.proxy_url),
            'cfg'      : copy.deepcopy(self._cfg)
        }
        return object_dict


    # --------------------------------------------------------------------------
    #
    @property
    def reg_addr(self):
        return self._cfg.reg_addr


    # --------------------------------------------------------------------------
    #
    @property
    def uid(self):
        return self._uid


    # --------------------------------------------------------------------------
    #
    @property
    def path(self):
        return self._cfg.path


    # --------------------------------------------------------------------------
    #
    @property
    def base(self):
        return self._cfg.base


    # --------------------------------------------------------------------------
    #
    @property
    def proxy_url(self):
        return self._cfg.proxy_url


    # --------------------------------------------------------------------------
    #
    @property
    def cfg(self):
        return self._cfg


    # --------------------------------------------------------------------------
    #
    @property
    def rcfgs(self):
        return self._rcfgs


    # --------------------------------------------------------------------------
    #
    @property
    def rcfg(self):
        return self._rcfg


    # --------------------------------------------------------------------------
    #
    @property
    def cmgr(self):
        return self._cmgr


    # --------------------------------------------------------------------------
    #
    def _get_logger(self, name, level=None, debug=None):
        """Get the Logger instance.

        This is a thin wrapper around `ru.Logger()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """
        log = ru.Logger(name=name, ns='radical.pilot', path=self._cfg.path,
                         targets=['.'], level=level, debug=debug)

        return log


    # --------------------------------------------------------------------------
    #
    def _get_reporter(self, name):
        """Get the Reporter instance.

        This is a thin wrapper around `ru.Reporter()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """

        if not self._reporter:

            enabled = ru.get_env_ns('report', 'radical.pilot', 'false').lower()

            if enabled in ['1', 'true', 'on']:
                enabled = True
            else:
                enabled = False

            self._reporter = ru.Reporter(name=name, ns='radical.pilot',
                                         path=self._cfg.path, enabled=enabled)
        return self._reporter


    # --------------------------------------------------------------------------
    #
    def _get_profiler(self, name):
        """Get the Profiler instance.

        This is a thin wrapper around `ru.Profiler()` which makes sure that
        log files end up in a separate directory with the name of `session.uid`.
        """

        prof = ru.Profiler(name=name, ns='radical.pilot', path=self._cfg.path)

        return prof


    # --------------------------------------------------------------------------
    #
    def inject_metadata(self, metadata):
        """Insert (experiment) metadata into an active session.

        RP stack version info always get added.
        """

        if not isinstance(metadata, dict):
            raise Exception("Session metadata should be a dict!")

        # FIXME MONGODB: to json
      # if self._dbs and self._dbs._c:
      #     self._dbs._c.update({'type'  : 'session',
      #                          "uid"   : self.uid},
      #                         {"$push" : {"metadata": metadata}})


    # --------------------------------------------------------------------------
    #
    def _register_pmgr(self, pmgr):

        self._pmgrs[pmgr.uid] = pmgr


  # # --------------------------------------------------------------------------
  # #
  # def _reconnect_pmgr(self, pmgr):
  #
  #     if not self._dbs.get_pmgrs(pmgr_ids=pmgr.uid):
  #         raise ValueError('could not reconnect to pmgr %s' % pmgr.uid)
  #
  #     self._pmgrs[pmgr.uid] = pmgr
  #
  #
    # --------------------------------------------------------------------------
    #
    def list_pilot_managers(self):
        """Get PilotManager instances.

        Lists the unique identifiers of all :class:`radical.pilot.PilotManager`
        instances associated with this session.

        Returns:
            list[str]: A list of :class:`radical.pilot.PilotManager` uids.

        """

        return list(self._pmgrs.keys())


    # --------------------------------------------------------------------------
    #
    def get_pilot_managers(self, pmgr_uids=None):
        """Get known PilotManager(s).

        Arguments:
            pmgr_uids (str | Iterable[str], optional): uids of the PilotManagers
                we want.

        Returns:
            radical.pilot.PilotManager | list[radical.pilot.PilotManager]: One
                or more `radical.pilot.PilotManager` objects.

        """

        return_scalar = False
        if isinstance(pmgr_uids, str):
            pmgr_uids     = [pmgr_uids]
            return_scalar = True

        if pmgr_uids: pmgrs = [self._pmgrs[uid] for uid in pmgr_uids]
        else        : pmgrs =  list(self._pmgrs.values())

        if return_scalar: return pmgrs[0]
        else            : return pmgrs


    # --------------------------------------------------------------------------
    #
    def _register_tmgr(self, tmgr):

        self._tmgrs[tmgr.uid] = tmgr


  # # --------------------------------------------------------------------------
  # #
  # def _reconnect_tmgr(self, tmgr):
  #
  #     if not self._dbs.get_tmgrs(tmgr_ids=tmgr.uid):
  #         raise ValueError('could not reconnect to tmgr %s' % tmgr.uid)
  #
  #     self._tmgrs[tmgr.uid] = tmgr
  #
  #
    # --------------------------------------------------------------------------
    #
    def list_task_managers(self):
        """Get TaskManager identifiers.

        Lists the unique identifiers of all :class:`radical.pilot.TaskManager`
        instances associated with this session.

        Returns:
            list[str]: A list of :class:`radical.pilot.TaskManager` uids.

        """

        return list(self._tmgrs.keys())


    # --------------------------------------------------------------------------
    #
    def get_task_managers(self, tmgr_uids=None):
        """Get known TaskManager(s).

        Arguments:
            tmgr_uids (str | list[str]): uids of the TaskManagers we want

        Returns:
            radical.pilot.TaskManager | list[radical.pilot.TaskManager]:
                One or more `radical.pilot.TaskManager` objects.

        """

        return_scalar = False
        if not isinstance(tmgr_uids, list):
            tmgr_uids     = [tmgr_uids]
            return_scalar = True

        if tmgr_uids: tmgrs = [self._tmgrs[uid] for uid in tmgr_uids]
        else        : tmgrs =  list(self._tmgrs.values())

        if return_scalar: return tmgrs[0]
        else            : return tmgrs


    # --------------------------------------------------------------------------
    #
    def list_resources(self):
        """Get list of known resource labels.

        Returns a list of known resource labels which can be used in a pilot
        description.
        """

        resources = list()
        for domain in self._rcfgs:
            for host in self._rcfgs[domain]:
                resources.append('%s.%s' % (domain, host))

        return sorted(resources)


    # --------------------------------------------------------------------------
    #
    def get_resource_config(self, resource, schema=None):
        """Returns a dictionary of the requested resource config."""

        site, res = resource.split('.', 1)
        if site not in self._rcfgs:
            raise RuntimeError("Resource site '%s' is unknown." % site)

        if res not in self._rcfgs[site]:
            raise RuntimeError("Resource label '%s' unknown." % res)

        if not schema:
            schema = self._rcfgs[site][res]['default_schema']

        if not schema:
            from_dict = self._rcfgs[site][res]
            from_dict.label = resource
            return ResourceConfig(from_dict=from_dict)

        if schema not in self._rcfgs[site][res]['schemas']:
            raise RuntimeError("schema %s unknown for resource %s"
                              % (schema, resource))

        rcfg = ResourceConfig(from_dict=self._rcfgs[site][res])
        scfg = rcfg['schemas'][schema]

        ru.dict_merge(rcfg, scfg, ru.OVERWRITE)

        if 'resource_manager' in rcfg:
            # import locally to avoid circular imports
            from .agent.resource_manager import ResourceManager

            rm = ResourceManager.get_manager(rcfg['resource_manager'])
            if rm and rm.batch_started():
                rcfg.update(ENDPOINTS_DEFAULT)

        rcfg.label = resource

        rcfg.verify()

        return rcfg



    # --------------------------------------------------------------------------
    #
    def fetch_profiles(self, tgt=None):

        return rpu.fetch_profiles(self._uid, tgt=tgt, skip_existing=True,
                                  rep=self._rep)


    # --------------------------------------------------------------------------
    #
    def fetch_logfiles(self, tgt=None):

        return rpu.fetch_logfiles(self._uid, tgt=tgt, skip_existing=True,
                                  rep=self._rep)


    # --------------------------------------------------------------------------
    #
    def _get_client_sandbox(self):
        """Client sandbox path.

        For the session in the client application, this is `os.getcwd()`.  For
        the session in any other component, specifically in pilot components,
        the client sandbox needs to be read from the session config (or pilot
        config).  The latter is not yet implemented, so the pilot can not yet
        interpret client sandboxes.  Since pilot-side staging to and from the
        client sandbox is not yet supported anyway, this seems acceptable
        (FIXME).
        """

        return self._cache['client_sandbox']


    # --------------------------------------------------------------------------
    #
    def _get_resource_sandbox(self, pilot):
        """Global RP sandbox.

        For a given pilot dict, determine the global RP sandbox, based on the
        pilot's 'resource' attribute.
        """

        # FIXME: this should get 'resource, schema=None' as parameters

        resource = pilot['description'].get('resource')
        schema   = pilot['description'].get('access_schema')

        if not resource:
            raise ValueError('Cannot get pilot sandbox w/o resource target')

        # the global sandbox will be the same for all pilots on any resource, so
        # we cache it
        with self._cache_lock:

            if resource not in self._cache['resource_sandbox']:

                # cache miss -- determine sandbox and fill cache
                rcfg   = self.get_resource_config(resource, schema)
                fs_url = ru.Url(rcfg['filesystem_endpoint'])

                # Get the sandbox from either the pilot_desc or resource conf
                sandbox_raw = pilot['description'].get('sandbox')
                if not sandbox_raw:
                    sandbox_raw = rcfg.get('default_remote_workdir', "$PWD")


                # we may need to replace pat elements with data from the pilot
                # description
                if '%' in sandbox_raw:
                    # expand from pilot description
                    expand = dict()
                    for k, v in pilot['description'].items():
                        if v is None:
                            v = ''
                        if k == 'project':
                            if '_' in v and 'ornl' in resource:
                                v = v.split('_')[0]
                            elif '-' in v and 'ncsa' in resource:
                                v = v.split('-')[0]
                        expand['pd.%s' % k] = v
                        if isinstance(v, str):
                            expand['pd.%s' % k.upper()] = v.upper()
                            expand['pd.%s' % k.lower()] = v.lower()
                        else:
                            expand['pd.%s' % k.upper()] = v
                            expand['pd.%s' % k.lower()] = v
                    sandbox_raw = sandbox_raw % expand


                # If the sandbox contains expandables, we need to resolve those
                # remotely.
                #
                # NOTE: this will only work for (gsi)ssh or similar shell
                #       based access mechanisms
                if '$' not in sandbox_raw:
                    # no need to expand further
                    sandbox_base = sandbox_raw

                else:
                    # FIXME "remote sandbox expansion not yet supported"
                    out, err, ret = ru.sh_callout(' echo "WORKDIR: %s"' %
                                                 sandbox_raw, shell=True)
                    if ret or 'WORKDIR:' not in out:
                        raise RuntimeError("workdir expansion failed: %s [%s]"
                                           % (out, err))

                    sandbox_base = out.split(":")[1].strip()
                    self._log.debug("sandbox base %s", sandbox_base)

                # at this point we have determined the remote 'pwd' - the
                # global sandbox is relative to it.
                fs_url.path = "%s/radical.pilot.sandbox" % sandbox_base

                # before returning, keep the URL string in cache
                self._cache['resource_sandbox'][resource] = fs_url

            return self._cache['resource_sandbox'][resource]


    # --------------------------------------------------------------------------
    #
    def _get_session_sandbox(self, pilot):

        # FIXME: this should get 'resource, schema=None' as parameters

        resource = pilot['description'].get('resource')

        if not resource:
            raise ValueError('Cannot get session sandbox w/o resource target')

        with self._cache_lock:

            if resource not in self._cache['session_sandbox']:

                # cache miss
                resource_sandbox      = self._get_resource_sandbox(pilot)
                session_sandbox       = ru.Url(resource_sandbox)
                session_sandbox.path += '/%s' % self.uid

                self._cache['session_sandbox'][resource] = session_sandbox

            return self._cache['session_sandbox'][resource]


    # --------------------------------------------------------------------------
    #
    def _get_pilot_sandbox(self, pilot):

        # FIXME: this should get 'pid, resource, schema=None' as parameters

        pilot_sandbox = pilot.get('pilot_sandbox')
        if str(pilot_sandbox):
            return ru.Url(pilot_sandbox)

        pid = pilot['uid']
        with self._cache_lock:

            if pid not in self._cache['pilot_sandbox']:

                # cache miss
                session_sandbox     = self._get_session_sandbox(pilot)
                pilot_sandbox       = ru.Url(session_sandbox)
                pilot_sandbox.path += '/%s/' % pilot['uid']

                self._cache['pilot_sandbox'][pid] = pilot_sandbox

            return self._cache['pilot_sandbox'][pid]


    # --------------------------------------------------------------------------
    #
    def _get_endpoint_fs(self, pilot):

        # FIXME: this should get 'resource, schema=None' as parameters

        resource = pilot['description'].get('resource')

        if not resource:
            raise ValueError("Can't get fs-endpoint w/o resource target")

        with self._cache_lock:

            if resource not in self._cache['endpoint_fs']:

                # cache miss
                resource_sandbox  = self._get_resource_sandbox(pilot)
                endpoint_fs       = ru.Url(resource_sandbox)
                endpoint_fs.path  = ''

                self._cache['endpoint_fs'][resource] = endpoint_fs

            return self._cache['endpoint_fs'][resource]


    # --------------------------------------------------------------------------
    #
    def _get_task_sandbox(self, task, pilot):

        # If a sandbox is specified in the task description, then interpret
        # relative paths as relativet to the pilot sandbox.

        # task sandboxes are cached in the task dict
        task_sandbox = task.get('task_sandbox')
        if task_sandbox:
            return task_sandbox

        # specified in description?
        if not task_sandbox:
            sandbox  = task['description'].get('sandbox')
            if sandbox:
                task_sandbox = ru.Url(self._get_pilot_sandbox(pilot))
                if sandbox[0] == '/':
                    task_sandbox.path = sandbox
                else:
                    task_sandbox.path += '/%s/' % sandbox

        # default
        if not task_sandbox:
            task_sandbox = ru.Url(self._get_pilot_sandbox(pilot))
            task_sandbox.path += "/%s/" % task['uid']

        # cache
        task['task_sandbox'] = str(task_sandbox)

        return task_sandbox


    # --------------------------------------------------------------------------
    #
    def _get_jsurl(self, pilot):
        """Get job service endpoint and hop URL for pilot's target resource."""

        resrc   = pilot['description']['resource']
        schema  = pilot['description']['access_schema']
        rcfg    = self.get_resource_config(resrc, schema)

        js_url  = ru.Url(rcfg.get('job_manager_endpoint'))
        js_hop  = ru.Url(rcfg.get('job_manager_hop') or js_url)

        # make sure the js_hop url points to an interactive access
        # TODO: this is an unreliable heuristics - we should require the js_hop
        #       URL to be specified in the resource configs.
        if   '+gsissh' in js_hop.schema or \
             'gsissh+' in js_hop.schema    : js_hop.schema = 'gsissh'
        elif '+ssh'    in js_hop.schema or \
             'ssh+'    in js_hop.schema    : js_hop.schema = 'ssh'
        else                               : js_hop.schema = 'fork'

        return js_url, js_hop


# ------------------------------------------------------------------------------





# pylint: disable=protected-access

__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru


# 'enum' for launch method types
LM_NAME_APRUN         = 'APRUN'
LM_NAME_CCMRUN        = 'CCMRUN'
LM_NAME_FORK          = 'FORK'
LM_NAME_IBRUN         = 'IBRUN'
LM_NAME_MPIEXEC       = 'MPIEXEC'
LM_NAME_MPIEXEC_MPT   = 'MPIEXEC_MPT'
LM_NAME_MPIRUN        = 'MPIRUN'
LM_NAME_MPIRUN_MPT    = 'MPIRUN_MPT'
LM_NAME_MPIRUN_CCMRUN = 'MPIRUN_CCMRUN'
LM_NAME_MPIRUN_DPLACE = 'MPIRUN_DPLACE'
LM_NAME_MPIRUN_RSH    = 'MPIRUN_RSH'
LM_NAME_JSRUN         = 'JSRUN'
LM_NAME_JSRUN_ERF     = 'JSRUN_ERF'
LM_NAME_PRTE          = 'PRTE'
LM_NAME_FLUX          = 'FLUX'
LM_NAME_RSH           = 'RSH'
LM_NAME_SSH           = 'SSH'
LM_NAME_SRUN          = 'SRUN'

PWD = os.getcwd()


# ------------------------------------------------------------------------------
#
class LaunchMethod(object):

    MPI_FLAVOR_OMPI     = 'OMPI'
    MPI_FLAVOR_HYDRA    = 'HYDRA'
    MPI_FLAVOR_SPECTRUM = 'SPECTRUM'
    MPI_FLAVOR_PALS     = 'PALS'
    MPI_FLAVOR_UNKNOWN  = 'unknown'


    # --------------------------------------------------------------------------
    #
    def __init__(self, name, lm_cfg, rm_info, log, prof):

        self.name      = name
        self._lm_cfg   = lm_cfg
        self._rm_info  = rm_info
        self._log      = log
        self._prof     = prof
        self._pwd      = os.getcwd()
        self._env_orig = ru.env_eval('env/bs0_orig.env')

        reg     = ru.zmq.RegistryClient(url=self._lm_cfg.reg_addr)
        lm_info = reg.get('lm.%s' % self.name.lower())

      # import pprint
      # self._log.debug('addr: %s', self._lm_cfg.reg_addr)
      # self._log.debug('name: %s', self.name)
      # self._log.debug('info: %s', pprint.pformat(lm_info))

        if not lm_info:

            # The registry does not yet contain any info for this LM - we need
            # to initialize the LM from scratch.  That happens in the env
            # defined by lm_cfg (derived from the original bs0 env)
            env_sh   = 'env/lm_%s.sh' % self.name.lower()
            env_lm   = ru.env_prep(environment=self._env_orig,
                          pre_exec=lm_cfg.get('pre_exec'),
                          pre_exec_cached=lm_cfg.get('pre_exec_cached'),
                          script_path=env_sh)

            # run init_from_scratch in a process under that derived env
            # FIXME: move this into init_from_scratch
            self._envp = ru.EnvProcess(env=env_lm)
            with self._envp:
                if self._envp:
                    data = self._init_from_scratch(env_lm, env_sh)
                    self._envp.put(data)
            lm_info = self._envp.get()

            # store the info in the registry for any other instances of the LM
            reg.put('lm.%s' % self.name.lower(), lm_info)
          # self._log.debug('INFO: %s', pprint.pformat(lm_info))

        reg.close()
        self._init_from_info(lm_info)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Launch Method.
    #
    @classmethod
    def create(cls, name, lm_cfg, rm_info, log, prof):

        # Make sure that we are the base-class!
        if cls != LaunchMethod:
            raise TypeError("LaunchMethod create only available to base class!")

        # In case of undefined LM just return None
        if not name:
            return None

        from .aprun          import APRun
        from .ccmrun         import CCMRun
        from .fork           import Fork
        from .ibrun          import IBRun
        from .mpiexec        import MPIExec
        from .mpirun         import MPIRun
        from .jsrun          import JSRUN
        from .prte           import PRTE
        from .flux           import Flux
        from .rsh            import RSH
        from .ssh            import SSH
        from .srun           import Srun

        try:
            impl = {
                LM_NAME_APRUN         : APRun,
                LM_NAME_CCMRUN        : CCMRun,
                LM_NAME_FORK          : Fork,
                LM_NAME_IBRUN         : IBRun,
                LM_NAME_MPIEXEC       : MPIExec,
                LM_NAME_MPIEXEC_MPT   : MPIExec,
                LM_NAME_MPIRUN        : MPIRun,
                LM_NAME_MPIRUN_CCMRUN : MPIRun,
                LM_NAME_MPIRUN_RSH    : MPIRun,
                LM_NAME_MPIRUN_MPT    : MPIRun,
                LM_NAME_MPIRUN_DPLACE : MPIRun,
                LM_NAME_JSRUN         : JSRUN,
                LM_NAME_JSRUN_ERF     : JSRUN,
                LM_NAME_PRTE          : PRTE,
                LM_NAME_FLUX          : Flux,
                LM_NAME_RSH           : RSH,
                LM_NAME_SSH           : SSH,
                LM_NAME_SRUN          : Srun,

            }

            if name not in impl:
                raise ValueError('LaunchMethod %s unknown' % name)

            return impl[name](name, lm_cfg, rm_info, log, prof)

        except Exception:
            log.exception('unusable lm %s' % name)
            raise


    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, env, env_sh):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def _init_from_info(self, lm_info):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        pass


    # --------------------------------------------------------------------------
    #
    def can_launch(self, task):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_launcher_env(self):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_task_named_env(self, env_name):

        # we assume that the launcher env is still active in the task execution
        # script.  We thus remove the launcher env from the task env before
        # applying the task env's pre_exec commands
        base = '%s/env/rp_named_env.%s' % (self._pwd, env_name)
        src  = '%s.env'                 %  base
        tgt  = '%s.%s.sh'               % (base, self.name.lower())

        # if the env does not yet exists - create
        # FIXME: this would need some file locking for concurrent executors. or
        #        add self._uid to path name
        if not os.path.isfile(tgt):
            ru.env_prep(environment=ru.env_read(src),
                        unset=list(os.environ.keys()),
                        script_path=tgt)

        return tgt


    # --------------------------------------------------------------------------
    #
    def get_launch_cmds(self, task, exec_path):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_rank_cmd(self):

        raise NotImplementedError("incomplete LaunchMethod %s" % self.name)


    # --------------------------------------------------------------------------
    #
    def get_exec(self, task):

        td           = task['description']
        task_exec    = td['executable']
        task_args    = td['arguments']
        task_argstr  = self._create_arg_string(task_args)
        command      = '%s %s' % (task_exec, task_argstr)

        return command.rstrip()


    # --------------------------------------------------------------------------
    #
    def get_partitions(self):

        return None


    # --------------------------------------------------------------------------
    #
    def _create_arg_string(self, args):

        if args:
            return ' '.join([ru.sh_quote(arg) for arg in args])
        else:
            return ''


    # --------------------------------------------------------------------------
    #
    def _get_mpi_info(self, exe):
        '''
        returns version and flavor of MPI version.
        '''

        version, flavor = None, None

        for info_opts in ['-V', '--version', '-info']:
            out, _, ret = ru.sh_callout('%s %s' % (exe, info_opts))
            if not ret:
                break

        if ret:
            raise RuntimeError('cannot identify MPI flavor [%s]' % exe)

        if 'pals' in exe.lower():
            flavor = self.MPI_FLAVOR_PALS

        for line in out.splitlines():

            if 'intel(r) mpi library for linux' in line.lower():
                # Intel MPI is hydra based
                version = line.split(',')[1].strip()
                flavor = self.MPI_FLAVOR_HYDRA

            elif 'hydra build details:' in line.lower():
                version = line.split(':', 1)[1].strip()
                flavor  = self.MPI_FLAVOR_HYDRA

            elif 'mvapich2' in line.lower():
                version = line.strip()
                flavor  = self.MPI_FLAVOR_HYDRA

            elif '(open mpi)' in line.lower():
                version = line.split(')', 1)[1].strip()
                flavor  = self.MPI_FLAVOR_OMPI

            elif 'ibm spectrum mpi' in line.lower():
                version = line.split(')', 1)[1].strip()
                flavor  = self.MPI_FLAVOR_SPECTRUM

            elif 'version' in line.lower():
                version = line.lower().split('version')[1].\
                          replace(':', '').strip()
                if not flavor:
                    flavor = self.MPI_FLAVOR_OMPI

            if version:
                break

        if not flavor:
            flavor = self.MPI_FLAVOR_UNKNOWN

        self._log.debug('mpi details [%s]: %s', exe, out)
        self._log.debug('mpi version: %s [%s]', version, flavor)

        return version, flavor


# ------------------------------------------------------------------------------






from typing import Any

import radical.utils as ru

# ------------------------------------------------------------------------------
#
class RPBaseMessage(ru.Message):

    # rpc distinguishes messages which are forwarded to the proxy bridge and
    # those which are not and thus remain local to the module they originate in.

    _schema   = {'fwd'      : bool}
    _defaults = {'_msg_type': 'rp_msg',
                 'fwd'      : False}


    # we do not register this message type - it is not supposed to be used
    # directly.


# ------------------------------------------------------------------------------
#
class HeartbeatMessage(RPBaseMessage):

    # heartbeat messages are never forwarded

    _schema   = {'uid'      : str}
    _defaults = {'_msg_type': 'heartbeat',
                 'fwd'      : False,
                 'uid'      : None}


ru.Message.register_msg_type('heartbeat', HeartbeatMessage)


# ------------------------------------------------------------------------------
#
class RPCRequestMessage(RPBaseMessage):

    _schema   = {'uid'      : str,   # uid of message
                 'addr'     : str,   # who is expected to act on the request
                 'cmd'      : str,   # rpc command
                 'args'     : list,  # rpc command arguments
                 'kwargs'   : dict}  # rpc command named arguments
    _defaults = {
                 '_msg_type': 'rpc_req',
                 'fwd'      : True,
                 'uid'      : None,
                 'addr'     : None,
                 'cmd'      : None,
                 'args'     : [],
                 'kwargs'   : {}}



ru.Message.register_msg_type('rpc_req', RPCRequestMessage)


# ------------------------------------------------------------------------------
#
class RPCResultMessage(RPBaseMessage):

    _schema   = {'uid'      : str,  # uid of rpc call
                 'val'      : Any,  # return value (`None` by default)
                 'out'      : str,  # stdout
                 'err'      : str,  # stderr
                 'exc'      : str}  # raised exception representation
    _defaults = {'_msg_type': 'rpc_res',
                 'fwd'      : True,
                 'uid'      : None,
                 'val'      : None,
                 'out'      : None,
                 'err'      : None,
                 'exc'      : None}

    # --------------------------------------------------------------------------
    #
    def __init__(self, rpc_req=None, from_dict=None, **kwargs):

        # when constructed from a request message copy the uid

        if rpc_req:
            if not from_dict:
                from_dict = dict()

            from_dict['uid'] = rpc_req['uid']

        super().__init__(from_dict, **kwargs)


ru.Message.register_msg_type('rpc_res', RPCResultMessage)


# ------------------------------------------------------------------------------





import os
import glob
import tarfile

import radical.utils as ru

from .. import constants as rpc

from .staging_helper import StagingHelper


# ------------------------------------------------------------------------------
#
def fetch_filetype(ext, name, sid, src=None, tgt=None, access=None,
        skip_existing=False, fetch_client=False, log=None, rep=None):
    '''
    Args:

        - ext  (str): file extension to fetch
        - name (str): full name of filetype for log messages etc
        - sid  (str): session for which all files are fetched
        - src  (str): dir to look for client session files (`$src/$sid/*.ext`)
        - tgt  (str): dir to store the files in (`$tgt/$sid/*.ext`,
            `$tgt/$sid/$pid/*.ext`)

    Returns:

        list[str]: list of file names (fetched and/or cached)

    '''

    if not log:
        log = ru.Logger('radical.pilot.utils')

    files = list()

    if not src:
        src = os.getcwd()

    if not tgt:
        tgt = os.getcwd()

    # no point in fetching client files into client sandbox
    if src == tgt:
        fetch_client = False

    if not tgt.startswith('/') and '://' not in tgt:
        tgt = "%s/%s" % (os.getcwd(), tgt)

    # we always create a session dir as real target
    tgt_url = ru.Url("%s/%s" % (tgt, sid))

    # turn URLs without `schema://host` into `file://localhost`,
    # so that they dont become interpreted as relative paths.
    if not tgt_url.schema: tgt_url.schema = 'file'
    if not tgt_url.host  : tgt_url.host   = 'localhost'

    # create target dir for session
    ru.rec_makedir(tgt_url.path)

    # first fetch client files
    if fetch_client:
        client_files = glob.glob("%s/%s/**.%s" % (src, sid, ext))
        if not client_files:
            raise RuntimeError('no client %s in %s/%s' % (name, src, sid))

        for client_file in client_files:

            ftgt = ru.Url('%s/%s' % (tgt_url, os.path.basename(client_file)))
            files.append("%s" % ftgt.path)

            if skip_existing and os.path.isfile(ftgt.path) \
                             and os.path.getsize(ftgt.path):
                pass
            else:
                log.debug('fetch client file %s', client_file)
                stager = StagingHelper(log)
                stager.copy(client_file, ftgt, flags=rpc.CREATE_PARENTS)

    # we need the session json for pilot details
    pilots = list()
    for fname in glob.glob('%s/pmgr.*.json' % sid):
        json_doc = ru.read_json(fname)
        pilots.extend(json_doc['pilots'])

    num_pilots = len(pilots)

    log.debug("Session: %s", sid)
    log.debug("Number of pilots in session: %d", num_pilots)

    for pilot in pilots:

        pid      = pilot['uid']
        tar_name = '%s.%s.tgz' % (pid, ext)

        # create target dir for this pilot
        ru.rec_makedir('%s/%s' % (tgt_url.path, pid))

        try:
            log.debug("processing pilot '%s'", pid)

            if access:
                # Allow to use a different access schema than used for the the
                # run.  Useful if you ran from the headnode, but would like to
                # retrieve the files to your desktop (Hello Titan).
                access_url = ru.Url(access)
                sandbox_url.schema = access_url.schema
                sandbox_url.host   = access_url.host

            else:
                sandbox_url = ru.Url(pilot['pilot_sandbox'])

            sandbox_url.path.rstrip('/')

            src_url  = ru.Url('%s/%s' % (sandbox_url, tar_name))
            src_dir  = os.path.dirname(src_url.path)
            is_local = False

            log.debug("sandbox: %s", sandbox_url)
            log.debug("src_url: %s", src_url)

            # check if we have a local tarball already
            if skip_existing and \
               os.path.isfile(tgt_url.path) and \
               os.path.getsize(tgt_url.path):
                is_local = True

            stager = StagingHelper(log)

            if not is_local:
                # need to fetch tarball
                #  - if no remote tarball exists, create it
                #  - fetch remote tarball
                _, _, ret = stager.sh_callout(src_url, 'test -f %s' % src_url.path)

                if ret:
                    # no tarball on remote side, create one
                    #
                    find_cmd = "find . -name '*.%s'" % ext
                    tar_cmd  = "cd %s && tar czf %s $(%s)" \
                             % (src_dir, tar_name, find_cmd)

                    out, err, ret = stager.sh_callout(src_url, tar_cmd)
                    log.debug("create with '%s': %s/%s", tar_cmd, out, err)

                    if ret:
                        raise RuntimeError("failed to create tarball: %s" % err)

                log.info("fetch '%s' to '%s'.", src_url, tgt_url)
                stager.copy(src_url, tgt_url, flags=rpc.CREATE_PARENTS)

            # we now have a local tarball - unpack it
            # note that we do not check if it was unpacked before - it's simpler
            # (and possibly cheaper) to just do that again
            log.info('Extract tarball %s', tgt_url.path)
            tarball = tarfile.open('%s/%s' % (tgt_url.path, tar_name), mode='r:gz')
            tarball.extractall("%s/%s" % (tgt_url.path, pid))

            found = glob.glob("%s/%s/*.%s" % (tgt_url.path, pid, ext))
            files.extend(found)

            if rep:
                rep.ok("+ %s (%s)\n" % (pid, name))

        except Exception:
            # do not raise, we still try the other pilots
            log.exception('failed to fetch %s for %s', pid, name)
            if rep:
                rep.error("- %s (%s)\n" % (pid, name))

    return files


# ------------------------------------------------------------------------------
#
def fetch_profiles (sid, src=None, tgt=None, access=None,
        skip_existing=False, fetch_client=False, log=None, rep=None):

    return fetch_filetype('prof', 'profiles', sid, src, tgt, access,
            skip_existing, fetch_client, log, rep)


# ------------------------------------------------------------------------------
#
def fetch_logfiles (sid, src=None, tgt=None, access=None,
        skip_existing=False, fetch_client=False, log=None, rep=None):

    return fetch_filetype('log', 'logfiles', sid, src, tgt, access,
            skip_existing, fetch_client, log, rep)


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2013-2020, http://radical.rutgers.edu'
__license__   = 'MIT'


from collections import defaultdict
from functools   import partial

import radical.utils as ru

from ...   import states as rps

from ..    import LaunchMethod
from ..    import ResourceManager
from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
#
class Flux(AgentExecutingComponent) :

    _shell = ru.which('bash') or '/bin/sh'

    # --------------------------------------------------------------------------
    #
    def initialize(self):
        '''
        This components has 3 strands of activity (threads):

          - the main thread listens for incoming tasks from the scheduler, and
            pushes them toward the watcher thread;
          - an event listener thread listens for flux events which signify task
            state updates, and pushes those events also to the watcher thread;
          - the watcher thread matches events and tasks, enacts state updates,
            and pushes completed tasks toward output staging.

        NOTE: we get tasks in *AGENT_SCHEDULING* state, and enact all
              further state changes in this component.
        '''

        super().initialize()

        # translate Flux states to RP states
        self._event_map = {'submit'   : None,   # rps.AGENT_SCHEDULING,
                           'depend'   : None,
                           'alloc'    : rps.AGENT_EXECUTING_PENDING,
                           'start'    : rps.AGENT_EXECUTING,
                           'cleanup'  : None,
                           'finish'   : rps.AGENT_STAGING_OUTPUT_PENDING,
                           'release'  : 'unschedule',
                           'free'     : None,
                           'clean'    : None,
                           'priority' : None,
                           'exception': rps.FAILED,
                          }

        # we get an instance of the resource manager (init from registry info)
        rm_name  = self.session.rcfg.resource_manager
        self._rm = ResourceManager.create(rm_name,
                                          self.session.cfg,
                                          self.session.rcfg,
                                          self._log, self._prof)

        lm_cfg  = self.session.rcfg.launch_methods.get('FLUX')
        lm_cfg['pid']       = self.session.cfg.pid
        lm_cfg['reg_addr']  = self.session.cfg.reg_addr
        self._lm            = LaunchMethod.create('FLUX', lm_cfg, self._rm.info,
                                                  self._log, self._prof)
        # local state management
        self._tasks      = defaultdict(dict)  # dict[partion_id][flux_id] = task
        self._task_count = 0


    # --------------------------------------------------------------------------
    #
    def cancel_task(self, uid):

        # FIXME: clarify how to cancel tasks in Flux
        pass


    # --------------------------------------------------------------------------
    #
    def _job_event_cb(self, partition_id, flux_id, event):

        task = self._tasks[partition_id].get(flux_id)
        if not task:
            self._log.error('no task for flux job %s: %s', flux_id, event.name)


        flux_state = event.name  # event: flux_id, flux_state
        state = self._event_map.get(flux_state)

        self._log.debug('flux event: %s: %s [%s]', flux_id, flux_state, state)

        if state is None:
            # ignore this state transition
            self._log.debug('ignore flux event %s:%s' %
                            (task['uid'], flux_state))

        # FIXME: how to get actual event transition timestamp?
        ts = event.timestamp

        self._log.debug('task state: %s [%s]', state, event)

        if state == rps.AGENT_STAGING_OUTPUT_PENDING:

            task['exit_code'] = event.context.get('status', 1)

            if task['exit_code']:
                task['target_state'] = rps.FAILED
            else:
                task['target_state'] = rps.DONE

            # on completion, push toward output staging
            self.advance_tasks(task, state, ts=ts, publish=True, push=True)

        elif state == 'unschedule':

            # free task resources
            self._prof.prof('unschedule_start', uid=task['uid'])
            self._prof.prof('unschedule_stop',  uid=task['uid'])  # ?
          # self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

        else:
            # otherwise only push a state update
            self.advance_tasks(task, state, ts=ts, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        # round robin on available flux partitions
        parts = defaultdict(list)
        for task in tasks:

            partition_id = task['description']['partition']

            if partition_id is None:
                partition_id = self._task_count % self._lm.n_partitions
                self._task_count += 1

            parts[partition_id].append(task)
            task['description']['environment']['RP_PARTITION_ID'] = partition_id

        for partition_id, partition_tasks in parts.items():

            part = self._lm.get_partition(partition_id)
            jds  = [self.task_to_spec(task) for task in partition_tasks]
            jids = part.submit_jobs([jd for jd in jds])
            self._log.debug('submitted tasks: %s', jids)

            for task, flux_id in zip(partition_tasks, jids):

                self._log.debug('submitted task %s -> %s', task['uid'], flux_id)

                md = task['description'].get('metadata') or dict()
                md['flux_id'] = flux_id
                task['description']['metadata'] = md

                self._tasks[partition_id][flux_id] = task

                event_cb = partial(self._job_event_cb, partition_id)

                part.attach_jobs([flux_id], event_cb)
                self._log.debug('handle %s: %s', task['uid'], flux_id)


    # --------------------------------------------------------------------------
    #
    def task_to_spec(self, task):

        td     = task['description']
        uid    = task['uid']
        sbox   = task['task_sandbox_path']
        stdout = td.get('stdout') or '%s/%s.out' % (sbox, uid)
        stderr = td.get('stderr') or '%s/%s.err' % (sbox, uid)

        task['stdout'] = ''
        task['stderr'] = ''

        task['stdout_file'] = stdout
        task['stderr_file'] = stderr

        _, exec_path = self._create_exec_script(self._lm, task)

        command = '%(cmd)s 1>%(out)s 2>%(err)s' % {'cmd': exec_path,
                                                   'out': stdout,
                                                   'err': stderr}

        spec = {
            'tasks': [{
                'slot' : 'task',
                'count': {
                    'per_slot': 1
                },
                'command': [self._shell, '-c', command],
            }],
            'attributes': {
                'system': {
                    'cwd'     : sbox,
                    'duration': 0,
                },
            },
            'version': 1,
            'resources': [{
                'count': td['ranks'],
                'type' : 'slot',
                'label': 'task',
                'with' : [{
                    'count': td['cores_per_rank'],
                    'type' : 'core'
              # }, {
              #     'count': int(td['gpus_per_rank'] or 0),
              #     'type' : 'gpu'
                }]
            }]
        }

        if td['gpus_per_rank']:

            gpr = td['gpus_per_rank']

            if gpr != int(gpr):
                raise ValueError('flux does not support on-integer GPU count')

            spec['resources'][0]['with'].append({
                    'count': int(gpr),
                    'type' : 'gpu'})

        return spec


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2023, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class Cobalt(ResourceManager):

    # --------------------------------------------------------------------------
    #
    @staticmethod
    def batch_started():

        return bool(os.getenv('COBALT_JOBID'))

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        if not rm_info.cores_per_node:
            raise RuntimeError('cores_per_node undetermined')

        if 'COBALT_NODEFILE' in os.environ:

            # this env variable is used for GPU nodes
            nodefile = os.environ['COBALT_NODEFILE']
            nodes    = self._parse_nodefile(nodefile, rm_info.cores_per_node)

        elif 'COBALT_PARTNAME' in os.environ:

            node_range = os.environ['COBALT_PARTNAME']
            nodes = [(node, rm_info.cores_per_node)
                     for node in ru.get_hostlist_by_range(node_range, 'nid', 5)]

            # Another option is to run `aprun` with the rank of nodes
            # we *think* we have, and with `-N 1` to place one rank per node,
            # and run `hostname` - that gives the list of hostnames.
            # (The number of nodes we receive from `$COBALT_PARTSIZE`.)
            #   out = ru.sh_callout('aprun -q -n %d -N 1 hostname' % n_nodes)[0]
            #   nodes = out.split()

        else:
            raise RuntimeError('no $COBALT_NODEFILE nor $COBALT_PARTNAME set')

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        return rm_info


# ------------------------------------------------------------------------------




#!/usr/bin/env python3
#dragon

USE_DRAGON = False

import os
import sys
import enum
import queue
import signal

if USE_DRAGON:
    import dragon                  # pylint: disable=unused-import, import-error

import multiprocessing as mp
import threading       as mt

import radical.utils   as ru
import radical.pilot   as rp

# ------------------------------------------------------------------------------
#
class Server(object):

    class Command(enum.Enum):
        TO_WATCH  = 1
        TO_CANCEL = 2


    # --------------------------------------------------------------------------
    #
    def __init__(self):

        sandbox   = sys.argv[1]
        self._log = ru.Logger('radical.pilot.dragon', path=sandbox)

        self._pin  = ru.zmq.Pipe(ru.zmq.MODE_PULL)
        self._pout = ru.zmq.Pipe(ru.zmq.MODE_PUSH)

        sys.stdout.write('ZMQ_ENDPOINTS %s %s\n'
                         % (ru.as_string(self._pin.url),
                            ru.as_string(self._pout.url)))
        sys.stdout.flush()

        self._watch_queue = queue.Queue()
        self._done_queue  = queue.Queue()

        self._watcher = mt.Thread(target=self._watch)
        self._watcher.daemon = True
        self._watcher.start()


    # --------------------------------------------------------------------------
    #
    def serve(self):

        # FIXME: profile events
        self._log.info('serving')

        if USE_DRAGON:
            mp.set_start_method("dragon")

        self._log.info('serving dragon')

        try:
            while True:

              # self._log.debug('main loop')

                self._handle_requests()
                self._handle_results()

        except Exception as e:
            self._log.exception('error in main loop: %s', e)
            raise


    # --------------------------------------------------------------------------
    #
    def _handle_requests(self):
        '''
        handle incoming requests from the parent process

        The requests are expected to be dictionaries with the following keys:
            - cmd:  the command to execute (run, cancel, stop)
            - task: the task to execute the command on
        '''

      # self._log.debug('waiting for request')

        msg = self._pin.get_nowait(0.1)

        if not msg:
            return

      # import pprint
      # self._log.debug('got request %s', pprint.pformat(msg))

        if not isinstance(msg, dict):
            self._log.error('invalid message type %s', type(msg))

        cmd  = msg.get('cmd')
        task = msg.get('task')

        if cmd not in ['run', 'cancel', 'stop']:
            self._log.error('unsupported command %s', cmd)
            return

        if cmd == 'stop':
            self._log.info('stopping')
            sys.exit(0)

        if not task:
            self._log.error('no task in message')
            return

        if cmd == 'cancel':
            self._log.debug('cancel task %s', task['uid'])
            self._watch_queue.put((self.Command.TO_CANCEL, task))
            return

        self._log.debug('launch task %s', task['uid'])
        task['proc'] = mp.Process(target=self._fork_task, args=[task, self._log])
        task['proc'].start()
        self._log.debug('task %s launched', task['uid'])

        # FIXME: this should be done in the watcher
        # self.handle_timeout(task)

        # watch task for completion
        self._watch_queue.put((self.Command.TO_WATCH, task))

        self._log.debug('task %s watched', task['uid'])


    # --------------------------------------------------------------------------
    #
    def _handle_results(self):

        if self._done_queue.empty():
            return

        task = self._done_queue.get()
        self._pout.put(task)


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def _fork_task(task, log):

        tid  = task['uid']
        sbox = task['task_sandbox_path']

        launch_script = '%s.launch.sh' % tid
        launch_out    = '%s/%s.launch.out' % (sbox, tid)
        launch_err    = '%s/%s.launch.err' % (sbox, tid)
        launch_cmd    = '%s/%s > %s 2> %s' % (sbox, launch_script,
                                                    launch_out, launch_err)
        out, err, ret = ru.sh_callout(launch_cmd, shell=True)

        log.debug('task %s completed %s : %s : %s' % (tid, out, err, ret))


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        to_watch  = list()
        to_cancel = list()

        # get completed tasks
        while True:

            if not self._watch_queue.empty():
                cmd, task = self._watch_queue.get()

                if cmd == self.Command.TO_WATCH:
                    to_watch.append(task)
                elif cmd == self.Command.TO_CANCEL:
                    to_cancel.append(task)
                else:
                    raise ValueError('unsupported cmd %s' % cmd)

            self._check_running(to_watch, to_cancel)


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self, to_watch, to_cancel):

        # `to_watch.remove()` in the loop requires copy to iterate over the list
        for task in list(to_watch):

            tid  = task['uid']
            proc = task['proc']

            tasks_to_advance = list()
            tasks_to_cancel  = list()

            if proc.is_alive():

                # process is still running - cancel if needed
                if tid in to_cancel:

                    # got a request to cancel this task - send SIGTERM to the
                    # process group (which should include the actual launch
                    # method)
                    try:
                        # kill the whole process group
                        proc.kill()

                    except OSError:
                        # lost race: task is already gone, we ignore this
                        # FIXME: collect and move to DONE/FAILED
                        pass

                    proc.join()

                    to_cancel.remove(tid)
                    to_watch.remove(task)
                    del task['proc']  # proc is not json serializable

                    tasks_to_cancel.append(task)

            else:

                # make sure proc is collected
                proc.join()
                task['exit_code'] = proc.exitcode

                # Free the Slots, Flee the Flots, Ree the Frots!
                to_watch.remove(task)
                if tid in to_cancel:
                    to_cancel.remove(tid)

                del task['proc']  # proc is not json serializable
                tasks_to_advance.append(task)

                self._log.debug('exit code %s: %s', task['uid'], proc.exitcode)

                if proc.exitcode != 0:
                    # task failed - fail after staging output
                    task['exception']        = 'RuntimeError("task failed")'
                    task['exception_detail'] = 'exit code: %s' % proc.exitcode
                    task['target_state'    ] = rp.FAILED

                else:
                    # The task finished cleanly, see if we need to deal with
                    # output data.  We always move to stageout, even if there
                    # are no directives -- at the very least, we'll upload
                    # stdout/stderr
                    task['target_state'] = rp.DONE

                self._done_queue.put({'cmd' : 'done',
                                      'task': task})


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    sys.stdout.write('started\n')
    sys.stdout.flush()

    s = Server()
    s.serve()


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import sys
import time

from collections import defaultdict

import radical.utils as ru
import radical.pilot as rp


def out(msg):
    sys.stdout.write('%s\n' % msg)
    sys.stdout.flush()


# This script has to run as a task within a pilot allocation, and is
# a demonstration of a task overlay within the RCT framework. It is expected
# to be staged and launched by the `raptor.py` script in the radical.pilot
# examples/misc directory.
# This master task will:
#
#   - create a master which bootstraps a specific communication layer
#   - insert n workers into the pilot (again as a task)
#   - perform RPC handshake with those workers
#   - send RPC requests to the workers
#   - terminate the worker
#
# The worker itself is an external program which is not covered in this code.

RANKS = 2


@rp.pythontask
def func_mpi(comm, msg, sleep):
    # pylint: disable=reimported
    import time
    print('hello %d/%d: %s' % (comm.rank, comm.size, msg))
    time.sleep(sleep)



@rp.pythontask
def func_non_mpi(a, sleep):
    # pylint: disable=reimported
    import math
    import random
    import time
    b = random.random()
    t = math.exp(a * b)
    time.sleep(sleep)
    print('hello! exp(%f) = %f' % (a * b, t))
    return t


# ------------------------------------------------------------------------------
#
class MyMaster(rp.raptor.Master):
    '''
    This class provides the communication setup for the task overlay: it will
    set up the request / response communication queues and provide the endpoint
    information to be forwarded to the workers.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg: ru.Config):

        self._cnt = 0
        self._submitted = defaultdict(int)
        self._collected = defaultdict(int)

        # initialize the task overlay base class.  That base class will ensure
        # proper communication channels to the pilot agent.
        super().__init__(cfg=cfg)
        self._sleep = cfg.sleep


    # --------------------------------------------------------------------------
    #
    def submit(self):

        self._prof.prof('create_start')

        # create additional tasks to be distributed to the workers.

        tds = list()
        for i in range(2):

            tds.append(rp.TaskDescription({
                'uid'        : 'task.exe.m.%06d' % i,
                'mode'       : rp.TASK_EXECUTABLE,
                'raptor_id'  : None,
                'ranks'      : RANKS,
                'executable' : '/bin/sh',
                'arguments'  : ['-c', 'sleep %d;' % self._sleep +
                             'echo "hello $RP_RANK/$RP_RANKS: $RP_TASK_ID"']}))

            tds.append(rp.TaskDescription({
                'uid'       : 'task.call.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_FUNC,
                'ranks'     : RANKS,
                'function'  : 'hello_mpi',
                'kwargs'          : {'msg': 'task.call.m.%06d' % i,
                                     'sleep': self._sleep},
                'raptor_id' : 'master.000000'}))

            bson = func_mpi(None, msg='task.call.m.%06d' % i, sleep=self._sleep)
            tds.append(rp.TaskDescription({
                'uid'       : 'task.mpi_ser_func.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_FUNC,
                'ranks'     : RANKS,
                'function'  : bson,
                'raptor_id' : 'master.000000'}))

            bson = func_non_mpi(i + 1, sleep=self._sleep)
            tds.append(rp.TaskDescription({
                'uid'       : 'task.ser_func.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_FUNC,
                'ranks'     : 1,
                'function'  : bson,
                'raptor_id' : 'master.000000'}))

            tds.append(rp.TaskDescription({
                'uid'       : 'task.eval.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_EVAL,
                'ranks'     : RANKS,
                'code'      :
                    'print("hello %%s/%%s: %%s" %% (os.environ["RP_RANK"],'
                    'os.environ["RP_RANKS"], os.environ["RP_TASK_ID"])) and '
                    'time.sleep(%d)' % self._sleep,
                'raptor_id' : 'master.000000'}))

            tds.append(rp.TaskDescription({
                'uid'       : 'task.exec.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_EXEC,
                'ranks'     : RANKS,
                'code'      :
                    'import time\ntime.sleep(%d)\n' % self._sleep +
                    'import os\nprint("hello %s/%s: %s" % (os.environ["RP_RANK"],'
                    'os.environ["RP_RANKS"], os.environ["RP_TASK_ID"]))',
                'raptor_id' : 'master.000000'}))

            tds.append(rp.TaskDescription({
                'uid'       : 'task.proc.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_PROC,
                'ranks'     : RANKS,
                'executable': '/bin/sh',
                'arguments' : ['-c',
                               'sleep %d; ' % self._sleep +
                               'echo "hello $RP_RANK/$RP_RANKS: '
                               '$RP_TASK_ID"'],
                'raptor_id' : 'master.000000'}))

            tds.append(rp.TaskDescription({
                'uid'       : 'task.shell.m.%06d' % i,
              # 'timeout'   : 10,
                'mode'      : rp.TASK_SHELL,
                'ranks'     : RANKS,
                'command'   : 'sleep %d; ' % self._sleep +
                              'echo "hello $RP_RANK/$RP_RANKS: $RP_TASK_ID"',
                'raptor_id' : 'master.000000'}))


        self.submit_tasks(tds)

        self._prof.prof('create_stop')

        # wait for outstanding tasks to complete
        while not self._term.is_set():

            completed = sum(self._collected.values())
            submitted = sum(self._submitted.values())

            if submitted:
                # request_cb has been called, so we can wait for completion

                self._log.info('submit done?: %d >= %d ', completed, submitted)

                if completed >= submitted:
                    break

            time.sleep(1)

        self._log.info('submit done!')


    # --------------------------------------------------------------------------
    #
    def request_cb(self, tasks):

        for task in tasks:

            self._log.debug('request_cb %s\n', task['uid'])

            mode = task['description']['mode']
            uid  = task['description']['uid']

            self._submitted[mode] += 1

            # for each `function` mode task, submit one more `proc` mode request
            if mode == rp.TASK_FUNC:
                self.submit_tasks(rp.TaskDescription(
                    {'uid'       : 'extra' + uid,
                   # 'timeout'   : 10,
                     'mode'      : rp.TASK_PROC,
                     'ranks'     : RANKS,
                     'executable': '/bin/sh',
                     'arguments' : ['-c',
                                    'sleep %d; ' % self._sleep +
                                    'echo "hello $RP_RANK/$RP_RANKS: '
                                          '$RP_TASK_ID"'],
                     'raptor_id' : 'master.000000'}))

        return tasks


    # --------------------------------------------------------------------------
    #
    def result_cb(self, tasks):
        '''
        Log results.

        Log file is named by the master tasks UID.
        '''
        for task in tasks:

            mode = task['description']['mode']
            self._collected[mode] += 1

            # NOTE: `state` will be `AGENT_EXECUTING`
            self._log.info('result_cb  %s: %s [%s] [%s]',
                            task['uid'],
                            task['state'],
                            task['stdout'],
                            task['return_value'])

            # Note that stdout is part of the master task result.
            print('id: %s [%s]:\n    out: %s\n    ret: %s\n'
                 % (task['uid'], task['state'], task['stdout'],
                    task['return_value']))


    # --------------------------------------------------------------------------
    #
    def worker_state_cb(self, worker_dict, state):

        if state == rp.AGENT_STAGING_OUTPUT_PENDING:
            self._log.warning('worker finalized')
            self.stop()


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    # This master script runs as a task within a pilot allocation.  The purpose
    # of this master is to (a) spawn a set or workers within the same
    # allocation, (b) to distribute work items (`hello` function calls) to those
    # workers, and (c) to collect the responses again.
    cfg_fname    = str(sys.argv[1])
    cfg          = ru.Config(cfg=ru.read_json(cfg_fname))
    cfg.rank     = int(sys.argv[2])

    n_workers        = cfg.n_workers
    nodes_per_worker = cfg.nodes_per_worker
    cores_per_node   = cfg.cores_per_node
    gpus_per_node    = cfg.gpus_per_node
    descr            = cfg.worker_descr

    # one node is used by master.  Alternatively (and probably better), we could
    # reduce one of the worker sizes by one core.  But it somewhat depends on
    # the worker type and application workload to judge if that makes sense, so
    # we leave it for now.

    # create a master class instance - this will establish communication to the
    # pilot agent
    master = MyMaster(cfg)

    # insert `n` worker tasks into the agent.  The agent will schedule (place)
    # those workers and execute them.  Insert one smaller worker (see above)
    # NOTE: this assumes a certain worker size / layout
    descr['ranks']         = nodes_per_worker * cores_per_node
    descr['gpus_per_rank'] = nodes_per_worker * gpus_per_node
    worker_ids = master.submit_workers(
                 [rp.TaskDescription(descr) for _ in range(n_workers)])

    # wait until `m` of those workers are up
    # This is optional, work requests can be submitted before and will wait in
    # a work queue.
    # FIXME
    master.wait_workers(count=1)

    master.start()
    master.submit()

    out('stop')
    # TODO: can be run from thread?
    master.stop()

    # TODO: worker state callback
    master.join()

    # TODO: expose RPC hooks


# ------------------------------------------------------------------------------

